import subprocess
import unittest
from collections import defaultdict
from itertools import groupby

import numpy as np
import pysam
from xopen import xopen

from celescope.tools import utils
from celescope.tools.step import Step, s_common


class Consensus(Step):
    """
    ## Features
    - Consensus all the reads of the same (barcode, UMI) combinations into one read(UMI). It will go through the sequence residue by residue and 
    count up the number of each type of residue (ie. A or G or T or C for DNA) in all sequences in the
    alignment. If the following conditions are met, the consensus sequence will be the most common residue in the alignment:
    1. the percentage of the most common residue type > threshold(default: 0.5);
    2. most common residue reads >= min_consensus_read;
    otherwise an ambiguous character(N) will be added.

    ## Output
    - `{sample}_consensus.fq` Fastq file after consensus.
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        # set
        self.min_consensus_read = int(self.args.min_consensus_read)

        # out files
        self.fq_tmp_file = f'{self.out_prefix}_sorted.fq.tmp'

        if self.args.out_fasta:
            self.consensus_out = f'{self.out_prefix}_consensus.fasta'
            self.out_fmt = "fasta"
        else:
            self.consensus_out = f'{self.out_prefix}_consensus.fq'
            self.out_fmt = "fastq"

    @utils.add_log
    def run(self):

        sort_fastq(self.args.fq, self.fq_tmp_file, self.outdir)
        n, total_ambiguous_base_n, length_list = sorted_dumb_consensus(
            fq=self.fq_tmp_file,
            outfile=self.consensus_out,
            threshold=self.args.threshold,
            min_consensus_read=self.min_consensus_read,
            out_fmt=self.out_fmt,
        )

        self.add_metric(
            name="UMI Counts",
            value=n,
            help_info='total UMI from FASTQ files',
        )
        self.add_metric(
            name="Mean UMI Length",
            value=round(np.mean(length_list), 2),
            help_info='mean of all UMI length'
        )
        self.add_metric(
            name="Ambiguous Base Counts",
            value=total_ambiguous_base_n,
            total=sum(length_list),
            help_info='number of bases that do not pass consensus threshold'
        )


@utils.add_log
def sort_fastq(fq, fq_tmp_file, outdir):
    tmp_dir = f'{outdir}/tmp'
    cmd = (
        f'mkdir {tmp_dir};'
        f'less {fq} | paste - - - - | sort -T {tmp_dir} -k1,1 -t " " | tr "\t" "\n" > {fq_tmp_file};'
    )
    subprocess.check_call(cmd, shell=True)


@utils.add_log
def sorted_dumb_consensus(fq, outfile, threshold, min_consensus_read, out_fmt):
    '''
    consensus read in name-sorted fastq
    output (barcode,umi) consensus fastq
    '''
    read_list = []
    n_umi = 0
    total_ambiguous_base_n = 0
    length_list = []
    out_h = xopen(outfile, 'w')

    def keyfunc(read):
        attr = read.name.split(':')
        return (attr[0], attr[1])

    with pysam.FastxFile(fq) as fh:
        for (barcode, umi), g in groupby(fh, key=keyfunc):
            read_list = []
            for read in g:
                read_list.append([read.sequence, read.quality])
            consensus_seq, consensus_qual, ambiguous_base_n, con_len = dumb_consensus(
                read_list,
                threshold=threshold,
                min_consensus_read=min_consensus_read,
                ambiguous="N"
            )
            n_umi += 1
            read_name = ':'.join([barcode, umi, str(n_umi)])
            if out_fmt == "fasta":
                out_h.write(utils.fasta_line(read_name, consensus_seq))
            else:
                out_h.write(utils.fastq_line(read_name, consensus_seq, consensus_qual))
            if n_umi % 10000 == 0:
                sorted_dumb_consensus.logger.info(f'{n_umi} UMI done.')
            total_ambiguous_base_n += ambiguous_base_n
            length_list.append(con_len)

    out_h.close()
    return n_umi, total_ambiguous_base_n, length_list


def dumb_consensus(read_list, threshold=0.5, min_consensus_read=1, ambiguous='N', default_qual='F'):
    '''
    This is similar to biopython dumb_consensus.
    It will just go through the sequence residue by residue and count up the number of each type
    of residue (ie. A or G or T or C for DNA) in all sequences in the
    alignment. If 
    1. the percentage of the most common residue type > threshold;
    2. most common residue reads >= min_consensus_read;
    then we will add that residue type,
    otherwise an ambiguous character will be added.
    elements of read_list: [entry.sequence,entry.quality]
    '''

    con_len = get_read_length(read_list, threshold=threshold)
    consensus_seq_list = []
    consensus_qual_list = []
    ambiguous_base_n = 0
    for n in range(con_len):
        atom_dict = defaultdict(int)
        quality_dict = defaultdict(int)
        num_atoms = 0
        for read in read_list:
            # make sure we haven't run past the end of any sequences
            # if they are of different lengths
            sequence = read[0]
            quality = read[1]
            if n < len(sequence):
                atom = sequence[n]
                atom_dict[atom] += 1
                num_atoms = num_atoms + 1

                base_qual = quality[n]
                quality_dict[base_qual] += 1

        consensus_atom = ambiguous
        for atom in atom_dict:
            if atom_dict[atom] > num_atoms * threshold and atom_dict[atom] >= min_consensus_read:
                consensus_atom = atom
                break
        if consensus_atom == ambiguous:
            ambiguous_base_n += 1
        consensus_seq_list.append(consensus_atom)

        max_freq_qual = 0
        consensus_base_qual = default_qual
        for base_qual in quality_dict:
            if quality_dict[base_qual] > max_freq_qual:
                max_freq_qual = quality_dict[base_qual]
                consensus_base_qual = base_qual

        consensus_qual_list.append(consensus_base_qual)
    
    consensus_seq = ''.join(consensus_seq_list)
    consensus_qual = ''.join(consensus_qual_list)

    return consensus_seq, consensus_qual, ambiguous_base_n, con_len


def get_read_length(read_list, threshold=0.5):
    '''
    compute read_length from read_list. 
    length = max length with read fraction >= threshold
    elements of read_list: [entry.sequence,entry.quality]
    '''

    n_read = len(read_list)
    length_dict = defaultdict(int)
    for read in read_list:
        length = len(read[0])
        length_dict[length] += 1
    for length in length_dict:
        length_dict[length] = length_dict[length] / n_read

    fraction = 0
    for length in sorted(length_dict.keys(), reverse=True):
        fraction += length_dict[length]
        if fraction >= threshold:
            return length


@utils.add_log
def consensus(args):
    if args.not_consensus:
        return
    with Consensus(args, display_title="Consensus") as runner:
        runner.run()


def get_opts_consensus(parser, sub_program):
    parser.add_argument("--threshold", help='Default 0.5. Valid base threshold. ', type=float, default=0.5)
    parser.add_argument("--not_consensus", help="Skip the consensus step. ", action='store_true')
    parser.add_argument("--min_consensus_read", help="Minimum number of reads to support a base. ", default=1)
    if sub_program:
        parser.add_argument("--fq", help="Required. Fastq file.", required=True)
        parser.add_argument("--out_fasta", default=False, help="output Fasta file or not", action='store_true')
        s_common(parser)


class Consensus_unittest(unittest.TestCase):

    def test_get_read_length(self):
        read_list = [['AAAA', 'FFFF'], ['TTT', 'FFF'], ['CCC', 'FFF'], ['GGGGGGG', 'FFFFFFF']]
        assert get_read_length(read_list, 0.5) == 4

    def test_dumb_consensus(self):
        read_list = [('AAAA', 'FFFF'), ('TTT', 'FF;'), ('CCCC', 'FFFF'), ('GGGAGGG', 'FFFFFFF')]
        consensus_seq, _consensus_qual, _ambiguous_base_n, _con_len = dumb_consensus(read_list, 0.5)
        self.assertEqual(consensus_seq, 'NNNA')

        read_list = [('AAAA', 'FFFF'), ('TTT', 'FF;'), ('CCC', 'FFF'), ('GGGGGGG', 'FFFFFFF')]
        consensus_seq, _consensus_qual, _ambiguous_base_n, _con_len = dumb_consensus(read_list, 0.5)
        self.assertEqual(consensus_seq, 'NNNN')


if __name__ == '__main__':
    unittest.main()



STEPS = [
    'sample',
    'starsolo',
    'analysis',
    'conversion',
    'substitution',
    'replacement']

__ASSAY__ = 'dynaseq'

IMPORT_DICT = {
    'analysis': 'celescope.rna',
}


#count
DYNA_MATRIX_DIR_SUFFIX = ['labeled','unlabeled']



# barcode
PATTERN_DICT = {
    'auto': None,
    'scopeV1': 'C12U8T18',
    'scopeV2.0.0': 'C8L16C8L16C8U8T18',
    'scopeV2.0.1': 'C8L16C8L16C8L1U8T18',
    'scopeV2.1.0': 'C8L16C8L16C8U12T18',
    'scopeV2.1.1': 'C8L16C8L16C8L1U12T18',
    'scopeV2.2.1': 'C8L16C8L16C8L1U12T18',
    'scopeV3.0.1': 'C9L16C9L16C9L1U12T18',
    # flv_rna is actually U9L16, but the last 10 bases can not be sequenced accurately.
    'flv_rna': 'C8L16C8L16C8U9L6',
    'flv' : 'U9C8L16C8L16C8',
    'bulk_vdj': 'L18C6U16',
    'bulk_rna': 'C9U12',
    'customized': None,
}


# count
OUTS_DIR = 'outs'
RAW_MATRIX_DIR_SUFFIX = 'raw'
FILTERED_MATRIX_DIR_SUFFIX = 'filtered'
MATRIX_FILE_NAME = 'matrix.mtx.gz'
FEATURE_FILE_NAME = 'features.tsv.gz'
BARCODE_FILE_NAME = 'barcodes.tsv.gz'
STAR_BAM_SUFFIX = 'Aligned.out.bam'
TAG_BAM_SUFFIX = 'aligned_posSorted_addTag.bam'
STARSOLO_BAM_SUFFIX = 'Aligned.sortedByCoord.out.bam'
COUNTS_FILE_NAME = 'counts.tsv'

# mkref
GENOME_CONFIG = 'celescope_genome.config'



"""
assign cell identity based on SNR and UMI_min
"""

from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.tools import utils
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt

import matplotlib

matplotlib.use('Agg')


def get_opts_count_tag(parser, sub_program):
    parser.add_argument(
        "--UMI_min",
        help="Default='auto'. Minimum UMI threshold. Cell barcodes with valid UMI < UMI_min are classified as *undeterminded*.",
        default="auto"
    )
    parser.add_argument(
        "--dim",
        help="Default=1. Tag dimentions. Usually we use 1-dimentional tag.",
        default=1
    )
    parser.add_argument(
        "--SNR_min",
        help="""Default='auto'. Minimum signal-to-noise ratio. 
Cell barcodes with UMI >=UMI_min and SNR < SNR_min are classified as *multiplet*. """,
        default="auto"
    )
    parser.add_argument("--combine_cluster",
                        help="Conbine cluster tsv file.", default=None)
    parser.add_argument(
        "--coefficient",
        help="""Default=0.1. If `SNR_min` is 'auto', minimum signal-to-noise ratio is calulated as 
`SNR_min = max(median(SNRs) * coefficient, 2)`. 
Smaller `coefficient` will cause less *multiplet* in the tag assignment.""",
        default=0.1
    )
    if sub_program:
        parser.add_argument("--read_count_file", help="Tag read count file.", required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--matrix_dir", help=HELP_DICT['matrix_dir'])
        parser.add_argument("--tsne_file", help=HELP_DICT['tsne_file'])

        s_common(parser)


def count_tag(args):

    with Count_tag(args, display_title="Cells") as runner:
        runner.run()


class Count_tag(Step):
    """
    ## Features
    - Assign tag to each cell barcode and summarize.

    ## Output

    - `{sample}_umi_tag.tsv` 

        `first column` cell barcode  
        `last column`  assigned tag  
        `columns between first and last` UMI count for each tag 

    - `{sample}_tsne_tag.tsv` it is `{sample}_umi_tag.tsv` with t-SNE coordinates, gene_counts and cluster infomation

    - `{sample}_cluster_count.tsv` cell barcode number assigned to *undeterminded*, *multiplet* and *each tag*

    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)
        self.read_count_file = args.read_count_file
        self.UMI_min = args.UMI_min
        self.SNR_min = args.SNR_min
        self.combine_cluster = args.combine_cluster
        self.dim = int(args.dim)
        self.coefficient = float(args.coefficient)

        # read
        self.df_read_count = pd.read_csv(self.read_count_file, sep="\t", index_col=0)

        if utils.check_arg_not_none(args, 'match_dir'):
            match_dict = utils.parse_match_dir(args.match_dir)
            self.match_barcode = match_dict['match_barcode']
            self.n_match_barcode = match_dict['n_match_barcode']
            self.tsne_file = match_dict['tsne_coord']
            self.matrix_dir = match_dict['matrix_dir']
        elif utils.check_arg_not_none(args, 'matrix_dir'):
            self.match_barcode, self.n_match_barcode = utils.get_barcode_from_matrix_dir(args.matrix_dir)
            self.tsne_file = args.tsne_file
            self.matrix_dir = args.matrix_dir
        else:
            raise ValueError("--match_dir or --matrix_dir is required.")

        # init
        self.no_noise = False

        # out files
        self.UMI_tag_file = f'{self.outdir}/{self.sample}_umi_tag.tsv'
        self.tsne_tag_file = f'{self.outdir}/{self.sample}_tsne_tag.tsv'
        self.cluster_count_file = f'{self.outdir}/{self.sample}_cluster_count.tsv'
        self.cluster_plot = f'{self.outdir}/{self.sample}_cluster_plot.pdf'
        if self.combine_cluster:
            self.combine_cluster_count_file = f'{self.outdir}/{self.sample}_combine_cluster_count.tsv'
            self.combine_cluster_plot = f'{self.outdir}/{self.sample}_combine_cluster_plot.pdf'

    @staticmethod
    def get_UMI(row):
        return row.sum()

    @staticmethod
    def get_UMI_min(df_cell_UMI, UMI_min):
        if UMI_min == "auto":
            UMI_min1 = np.percentile(df_cell_UMI.sum(axis=1), 5)
            UMI_min2 = np.median(df_cell_UMI.sum(axis=1)) / 10
            UMI_min = int(min(UMI_min1, UMI_min2))
            UMI_min = max(UMI_min, 1)
            return UMI_min
        else:
            return int(UMI_min)

    @staticmethod
    def get_SNR(row, dim):
        row_sorted = sorted(row, reverse=True)
        noise = row_sorted[dim]
        signal = row_sorted[dim - 1]
        if signal == 0:
            return 0
        if noise == 0:
            return np.inf
        return float(signal) / noise

    @utils.add_log
    def get_SNR_min(self, df_cell_UMI, SNR_min, UMI_min):
        UMIs = df_cell_UMI.apply(Count_tag.get_UMI, axis=1)
        df_valid_cell_UMI = df_cell_UMI[UMIs >= UMI_min]
        if SNR_min == "auto":
            # no noise
            if df_valid_cell_UMI.shape[1] <= self.dim:
                Count_tag.get_SNR_min.logger.warning('*** No NOISE FOUND! ***')
                self.no_noise = True
                return 0
            SNRs = df_valid_cell_UMI.apply(Count_tag.get_SNR, dim=self.dim, axis=1)
            if np.median(SNRs) == np.inf:
                return 10
            return max(np.median(SNRs) * self.coefficient, 2)
        else:
            return float(SNR_min)

    @staticmethod
    def tag_type(row, UMI_min, SNR_min, dim, no_noise=False):
        if no_noise:
            SNR = 1
        else:
            SNR = Count_tag.get_SNR(row, dim)
        UMI = Count_tag.get_UMI(row)
        if UMI < UMI_min:
            return "Undetermined"
        if SNR < SNR_min:
            return "Multiplet"
        # get tag
        signal_tags = sorted(row.sort_values(ascending=False).index[0:dim])
        signal_tags_str = "_".join(signal_tags)
        return signal_tags_str

    @utils.add_log
    def write_and_plot(self, df, column_name, count_file, plot_file):
        df_count = df.groupby(["tag", column_name]).size().unstack()
        df_count.fillna(0, inplace=True)
        df_count.to_csv(count_file, sep="\t")
        df_percent = df_count / df_count.sum()
        df_plot = df_percent.stack().reset_index()
        df_plot.rename({0: "percent"}, axis=1, inplace=True)

        # plot
        colors = list(matplotlib.colors.cnames.keys())
        fig, ax = plt.subplots(figsize=(20, 10))
        types = df_plot["tag"].drop_duplicates()
        margin_bottom = np.zeros(len(df_plot[column_name].drop_duplicates()))

        for num, tag_type in enumerate(types):
            try:
                color = colors[num * 3 + 1]
            except IndexError:
                print("not enough colors")
                return
            values = list(df_plot.loc[df_plot["tag"] == tag_type, "percent"])
            df_plot[df_plot['tag'] == tag_type].plot.bar(
                x=column_name, y='percent', ax=ax, stacked=True,
                bottom=margin_bottom, label=tag_type, color=color)

            margin_bottom += values
        ax.legend(loc='center left', bbox_to_anchor=(1, 0.5))
        plt.title("tag fraction")
        fig.savefig(plot_file)

    @utils.add_log
    def run(self):

        mapped_read = int(self.df_read_count['read_count'].sum())

        # in cell
        df_read_count_in_cell = self.df_read_count[self.df_read_count.index.isin(self.match_barcode)]
        mapped_read_in_cell = int(df_read_count_in_cell['read_count'].sum())
        self.add_metric(
            name='Mapped Reads in Cells',
            value=mapped_read_in_cell,
            total=mapped_read,
            help_info="Mapped reads with scRNA-Seq cell barcode"
        )

        # UMI
        tag_name = df_read_count_in_cell.columns[0]
        df_UMI_in_cell = df_read_count_in_cell.reset_index().groupby([
            'barcode', tag_name]).agg({'UMI': 'count'})
        df_UMI_in_cell = df_UMI_in_cell.reset_index()
        df_UMI_in_cell = df_UMI_in_cell.pivot(
            index='barcode', columns=tag_name, values='UMI')
        df_cell = pd.DataFrame(index=self.match_barcode)
        df_UMI_cell = pd.merge(
            df_cell,
            df_UMI_in_cell,
            how="left",
            left_index=True,
            right_index=True
        )

        # fillna
        df_UMI_cell.fillna(0, inplace=True)
        df_UMI_cell = df_UMI_cell.astype(int)

        # UMI
        UMIs = df_UMI_cell.apply(sum, axis=1)
        umi_median = round(np.median(UMIs), 2)
        umi_mean = round(np.mean(UMIs), 2)
        self.add_metric(
            name='Median UMI per Cell',
            value=umi_median,
            help_info="Median UMI per scRNA-Seq cell barcode"
        )

        self.add_metric(
            name='Mean UMI per Cell',
            value=umi_mean,
            help_info="Mean UMI per scRNA-Seq cell barcode"
        )

        UMI_min = Count_tag.get_UMI_min(df_UMI_cell, self.UMI_min)
        Count_tag.run.logger.info(f'UMI_min: {UMI_min}')
        SNR_min = self.get_SNR_min(df_UMI_cell, self.SNR_min, UMI_min)
        Count_tag.run.logger.info(f'SNR_min: {SNR_min}')
        df_UMI_cell["tag"] = df_UMI_cell.apply(
            Count_tag.tag_type, UMI_min=UMI_min, SNR_min=SNR_min, dim=self.dim, no_noise=self.no_noise, axis=1)
        df_UMI_cell.to_csv(self.UMI_tag_file, sep="\t")

        df_tsne = pd.read_csv(self.tsne_file, sep="\t", index_col=0)
        df_tsne_tag = pd.merge(
            df_tsne,
            df_UMI_cell,
            how="left",
            left_index=True,
            right_index=True)

        if self.combine_cluster:
            df_combine_cluster = pd.read_csv(
                self.combine_cluster, sep="\t", header=None)
            df_combine_cluster.columns = ["cluster", "combine_cluster"]
            df_tsne_combine_cluster_tag = pd.merge(
                df_tsne_tag, df_combine_cluster,
                on=["cluster"], how="left", left_index=True).set_index(df_tsne_tag.index)
            df_tsne_combine_cluster_tag.to_csv(self.tsne_tag_file, sep="\t")
        else:
            df_tsne_tag.to_csv(self.tsne_tag_file, sep="\t")

        self.write_and_plot(
            df=df_tsne_tag,
            column_name="cluster",
            count_file=self.cluster_count_file,
            plot_file=self.cluster_plot,
        )

        if self.combine_cluster:
            self.write_and_plot(
                df=df_tsne_combine_cluster_tag,
                column_name="combine_cluster",
                count_file=self.combine_cluster_count_file,
                plot_file=self.combine_cluster_plot
            )

        sr_tag_count = df_UMI_cell["tag"].value_counts()  # series(index:tag name, value:tag count)
        for tag_name in ("Undetermined", "Multiplet"):
            if tag_name in sr_tag_count:
                self.add_metric(
                    name=tag_name + ' Cells',
                    value=int(sr_tag_count[tag_name]),
                    total=self.n_match_barcode,
                )
                sr_tag_count.drop(tag_name, inplace=True)
        for tag_name in sorted(sr_tag_count.index):
            self.add_metric(
                name=tag_name + ' Cells',
                value=int(sr_tag_count[tag_name]),
                total=self.n_match_barcode,
            )




import pandas as pd
import pysam

from celescope.tools import utils
from celescope.tools.barcode import Barcode
from celescope.tools.step import Step, s_common

# n_mismatch = 1 if n_tag_barcode > N_TAG_BARCODE_THRESHOLD else 2
N_TAG_BARCODE_THRESHOLD = 10000

def get_opts_mapping_tag(parser, sub_program):
    parser.add_argument(
        "--fq_pattern",
        help="""R2 read pattern. The number after the letter represents the number of bases. The `fq_pattern` of CLindex is `L25C15`
`L` linker(common sequences)  
`C` tag barcode  
""",
        default='L25C15'
    )
    parser.add_argument(
        "--barcode_fasta",
        help="""Required. Tag barcode fasta file. It will check the mismatches between tag barcode 
sequence in R2 reads with all tag barcode sequence in barcode_fasta. 
It will assign read to the tag with mismatch < threshold. 
If no such tag exists, the read is classified as invalid.

You can find the barcode fasta file under `celescope/data/Clindex`
```
>CLindex_TAG_1
CGTGTTAGGGCCGAT
>CLindex_TAG_2
GAGTGGTTGCGCCAT
>CLindex_TAG_3
AAGTTGCCAAGGGCC
>CLindex_TAG_4
TAAGAGCCCGGCAAG
>CLindex_TAG_5
TGACCTGCTTCACGC
>CLindex_TAG_6
GAGACCCGTGGAATC
>CLindex_TAG_7
GTTATGCGACCGCGA
>CLindex_TAG_8
ATACGCAGGGTCCGA
>CLindex_TAG_9
AGCGGCATTTGGGAC
>CLindex_TAG_10
TCGCCAGCCAAGTCT
>CLindex_TAG_11
ACCAATGGCGCATGG
>CLindex_TAG_12
TCCTCCTAGCAACCC
>CLindex_TAG_13
GGCCGATACTTCAGC
>CLindex_TAG_14
CCGTTCGACTTGGTG
>CLindex_TAG_15
CGCAAGACACTCCAC
>CLindex_TAG_16
CTGCAACAAGGTCGC
```
""",
        required=True,
    )
    parser.add_argument(
        "--linker_fasta",
        help="""Optional. If provided, it will check the mismatches between linker sequence in R2 reads 
with all linker sequence in linker_fasta. If no mismatch < len(linker) / 10 + 1, the read is classified as invalid.
""",
    )
    if sub_program:
        s_common(parser)
        parser.add_argument("--fq", help="R2 read fastq.", required=True)


@utils.add_log
def mapping_tag(args):
    with Mapping_tag(args, display_title="Mapping") as runner:
        runner.run()


class Mapping_tag(Step):
    """
    ## Features
    - Align R2 reads to the tag barcode fasta.

    ## Output

    - `{sample}_read_count.tsv` tab-delimited text file with 4 columns.

        `barcode` cell barcode  
        `tag_name`  tag name in barcode_fasta  
        `UMI`   UMI sequence  
        `read_count` read count per UMI

    - `{sample}_invalid_barcode.tsv` tab-delimited text file with 2 columns.
        `tag_barcode` tag barcodes that do not match with any sequence in `--barcode_fasta`.
        `read_count` invalid tag barcode read counts
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # read args
        self.fq = args.fq
        self.fq_pattern = args.fq_pattern
        self.linker_fasta = args.linker_fasta
        self.barcode_fasta = args.barcode_fasta

        # process
        self.pattern_dict = Barcode.parse_pattern(self.fq_pattern)

        self.barcode_dict, self.barcode_length = utils.read_fasta(self.barcode_fasta, equal=True)
        len_C = Barcode.get_abbr_len(self.pattern_dict, 'C')
        if len_C != self.barcode_length:
            raise ValueError(f"""The length of tag barcode in fq_pattern({len_C}) != 
                length of tag barcode in barcode_fasta({self.barcode_length})""")

        if self.linker_fasta and self.linker_fasta != 'None':
            self.linker_dict, self.linker_length = utils.read_fasta(self.linker_fasta, equal=True)
            len_L = Barcode.get_abbr_len(self.pattern_dict, 'L')
            if len_L != self.linker_length:
                raise ValueError(f"""The length of linker in fq_pattern({len_L}) != 
                    length of linker in linker_fasta({self.linker_length})""")
        else:
            self.linker_dict, self.linker_length = {}, 0


        # mismatch
        self.mismatch_dict = self.get_tag_barcode_mismatch_dict()

        # variables
        self.total_reads = 0
        self.reads_unmapped_too_short = 0
        self.reads_unmapped_invalid_linker = 0
        self.reads_unmapped_invalid_barcode = 0
        self.reads_mapped = 0
        self.res_dic = utils.genDict()
        self.res_sum_dic = utils.genDict(dim=2)
        self.match_barcode = []
        self.invalid_barcode_dict = utils.genDict(dim=1)

        # out files
        self.read_count_file = f'{self.outdir}/{self.sample}_read_count.tsv'
        self.invalid_barcode_file = f'{self.outdir}/{self.sample}_invalid_barcode.tsv'

    @utils.add_log
    def get_tag_barcode_mismatch_dict(self):
        mismatch_dict = {}
        n_mismatch = 1 if len(self.barcode_dict) > N_TAG_BARCODE_THRESHOLD else 2
        for seq_id, seq in self.barcode_dict.items():
            for mismatch_seq in Barcode.findall_mismatch(seq, n_mismatch=n_mismatch):
                mismatch_dict[mismatch_seq] = seq_id   
        
        return mismatch_dict


    def check_barcode_with_mismatch(self, barcode, seq_barcode, umi):
        """
        Args:
            barcode: cell barcode
            seq_barcode: tag barcode sequence
            umi: UMI sequence
        """
        if seq_barcode in self.mismatch_dict:
            seq_id = self.mismatch_dict[seq_barcode]
            self.res_dic[barcode][seq_id][umi] += 1
            self.reads_mapped += 1
        else:
            self.reads_unmapped_invalid_barcode += 1
            self.invalid_barcode_dict[seq_barcode] += 1

    def process_read(self):

        with pysam.FastxFile(self.fq) as infile:
            for record in infile:
                self.total_reads += 1
                attr = str(record.name).strip("@").split(':')
                barcode = str(attr[0])
                umi = str(attr[1])
                seq = record.sequence

                if self.linker_length != 0:
                    seq_linker = Barcode.get_seq_str_no_exception(seq, self.pattern_dict['L'])
                    if len(seq_linker) < self.linker_length:
                        self.reads_unmapped_too_short += 1
                        continue
                if self.barcode_dict:
                    seq_barcode = Barcode.get_seq_str_no_exception(seq, self.pattern_dict['C'])
                    if self.barcode_length != len(seq_barcode):
                        miss_length = self.barcode_length - len(seq_barcode)
                        if miss_length > 2:
                            self.reads_unmapped_too_short += 1
                            continue
                        seq_barcode = seq_barcode + "A" * miss_length

                # check linker
                if self.linker_length != 0:
                    valid_linker = False
                    for linker_name in self.linker_dict:
                        if utils.hamming_correct(self.linker_dict[linker_name], seq_linker):
                            valid_linker = True
                            break
                else:
                    valid_linker = True

                if not valid_linker:
                    self.reads_unmapped_invalid_linker += 1
                    continue

                # check barcode
                self.check_barcode_with_mismatch(barcode, seq_barcode, umi)

    def write_files(self):
        # write dic to pandas df
        rows = []
        for barcode in self.res_dic:
            for tag_name in self.res_dic[barcode]:
                for umi in self.res_dic[barcode][tag_name]:
                    rows.append([barcode, tag_name, umi,
                                 self.res_dic[barcode][tag_name][umi]])
        df_read_count = pd.DataFrame(rows)
        df_read_count.rename(
            columns={
                0: "barcode",
                1: "tag_name",
                2: "UMI",
                3: "read_count"
            }, inplace=True)
        df_read_count.to_csv(
            self.read_count_file, sep='\t', index=False)

        # write invalid seq_barcode to file
        with open(self.invalid_barcode_file, "w") as f:
            f.write("tag_barcode\tread_count\n")
            for seq_barcode,count in sorted(self.invalid_barcode_dict.items(), key=lambda x: x[1], reverse=True):
                f.write(f"{seq_barcode}\t{count}\n")

    def add_metrics(self):
        # add metrics
        self.add_metric(
            name='Reads Mapped',
            value=self.reads_mapped,
            total=self.total_reads,
            help_info="R2 reads that successfully mapped to linker and tag-barcode"
        )
        self.add_metric(
            name='Reads Unmapped too Short',
            value=self.reads_unmapped_too_short,
            total=self.total_reads,
            help_info="Unmapped R2 reads because read length < linker length + tag-barcode length"
        )
        self.add_metric(
            name='Reads Unmapped Invalid Linker',
            value=self.reads_unmapped_invalid_linker,
            total=self.total_reads,
            help_info="Unmapped R2 reads because of too many mismatches in linker sequence"
        )
        self.add_metric(
            name='Reads Unmapped Invalid Barcode',
            value=self.reads_unmapped_invalid_barcode,
            total=self.total_reads,
            help_info="Unmapped R2 reads because of too many mismatches in tag-barcode sequence"
        )

    @utils.add_log
    def run(self):
        self.process_read()
        self.write_files()
        self.add_metrics()



from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.rna.mkref import Mkref_rna
from celescope.snp.__init__ import PANEL

class Variant_calling(Step):
    """
    ## Features
    - Perform variant calling at single cell level.

    ## Output
    - `{sample}_raw.vcf` Variants are called with bcftools default settings.
    - `{sample}_norm.vcf` Indels are left-aligned and normalized. See https://samtools.github.io/bcftools/bcftools.html#norm for more details.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # set
        self.barcodes, _num = utils.get_barcode_from_match_dir(args.match_dir)
        self.fasta = Mkref_rna.get_config(args.genomeDir)['files']['fasta']
        self.df_vcf = None
        self.panel = args.panel
        self.bed = utils.get_bed_file_path(self.panel)

        # out
        self.splitN_bam = f'{self.out_prefix}_splitN.bam'
        self.splitN_bam_name_sorted = f'{self.out_prefix}_splitN_name_sorted.bam'

        self.raw_bcf_file = f'{self.out_prefix}_raw.bcf'
        self.raw_vcf_file = f'{self.out_prefix}_raw.vcf'
        self.fixed_header_vcf = f'{self.out_prefix}_fixed.vcf'
        self.norm_vcf_file = f'{self.out_prefix}_norm.vcf'

    @utils.add_log
    def SplitNCigarReads(self):
        cmd = (
            f'gatk '
            f'SplitNCigarReads '
            f'--do-not-fix-overhangs '
            f'-R {self.fasta} '
            f'-I {self.args.bam} '
            f'-O {self.splitN_bam} '
        )
        self.debug_subprocess_call(cmd)

    @utils.add_log
    def gatk_norm(self):
        cmd = (
            f'gatk LeftAlignAndTrimVariants '
            f'-R {self.fasta} '
            f'-V {self.fixed_header_vcf} '
            f'-O {self.norm_vcf_file} '
            '--split-multi-allelics '
        )
        self.debug_subprocess_call(cmd)

    @utils.add_log
    def call_variants(self):
        """
        max depth 100M
        """
        cmd = (
            f'bcftools mpileup '
            f'-f {self.fasta} '
            f'--threads {self.thread} '
            f'--annotate DP,AD -d 100000000 '
            f'-o {self.raw_bcf_file} '
            f'{self.splitN_bam} '
        )
        if self.bed:
            cmd += f' --regions-file {self.bed} '
        self.debug_subprocess_call(cmd)

        cmd = (
            f'bcftools call '
            f'-mv -Ov '
            f'-o {self.raw_vcf_file} '
            f'{self.raw_bcf_file} '
        )
        self.debug_subprocess_call(cmd)

    def bcftools_norm(self):
        cmd = (
            'bcftools norm '
            '-m- '
            f'-f {self.fasta} '
            f'{self.raw_vcf_file} '
            '| bcftools norm '
            '-d both '
            f'-o {self.norm_vcf_file} '
        )
        self.debug_subprocess_call(cmd)

    def run(self):

        self.SplitNCigarReads()
        self.call_variants()
        self.bcftools_norm()

    


@utils.add_log
def variant_calling(args):

    with Variant_calling(args) as runner:
        runner.run()


def get_opts_variant_calling(parser, sub_program):

    parser.add_argument("--genomeDir", help=HELP_DICT['genomeDir'], required=True)
    parser.add_argument("--panel", help=HELP_DICT['panel'], choices=list(PANEL))
    if sub_program:
        parser.add_argument(
            "--bam",
            help='Input BAM file from step `target_metrics`. ',
            required=True
        )
        parser.add_argument(
            "--match_dir",
            help=HELP_DICT['match_dir'],
            required=True
        )
        s_common(parser)



import pysam
import os
import subprocess
import pandas as pd
from collections import defaultdict
from multiprocessing import Pool

from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.rna.mkref import Mkref_rna
from celescope.tools import reference
from celescope.__init__ import HELP_DICT
from celescope.tools.plotly_plot import Conversion_plot


class Conversion(Step):
    """
    ## Features
    - Get conversion pos in each read.
        - Get snp info. 

    ## Output
    - `{sample}.PosTag.bam` Bam file with conversion info.
    - `{sample}.PosTag.csv` TC conversion sites info in csv format.
    - `{sample}.snp.csv` Candidated snp sites.
    """

    def __init__(self, args):
        Step.__init__(self, args)
        # input files
        self.sample = args.sample
        self.inbam = args.bam
        self.bcfile = args.cell
        self.outdir = args.outdir
        self.thread = int(args.thread)
        self.qual = int(args.basequalilty)
        self.snp_min_cells = args.snp_min_cells
        self.snp_min_depth = args.snp_min_depth
        self.cellsplit = args.cellsplit

        # set
        gtf_file = Mkref_rna.get_config(args.genomeDir)['files']['gtf']
        gp = reference.GtfParser(gtf_file)
        gp.get_id_name()
        self.strand_dict = gp.get_strand()
        self.cell_dict, self.cell_num = utils.barcode_list_stamp(self.bcfile,cut=self.cellsplit)
        self.bam_list = []
        self.conv_df = pd.DataFrame()
        self.snp_df = pd.DataFrame()
        

        # output files 
        ## tmp outputdir
        self.tmp_dir = f'{args.outdir}/tmp'
        utils.check_mkdir(self.tmp_dir)

        ## final outputs
        self.outfile_bam = os.path.join(args.outdir, args.sample+'.PosTag.bam')
        self.outtmp_bam = os.path.join(args.outdir, args.sample+'.PosTag.bam.tmp.bam')
        self.outfile_csv = os.path.join(args.outdir, args.sample+'.PosTag.csv')
        self.outsnp_csv = os.path.join(args.outdir, args.sample+'.snp.csv')

    @utils.add_log
    def run(self):

        # Adding tags and parse snps
        dfs = self.run_conversion()
        # Obtaining conversion positions
        self.snp_candidate(dfs)
        # merge bam files
        self.output_bam()
        # stat and plot
        self.add_conversion_metrics()
        self.conversion_plot()
        # delete tmp dir
        self.clean_tmp()



    def run_cmd(self, cmd):
        subprocess.call(' '.join(cmd), shell=True)

    @utils.add_log
    def run_conversion(self):
        cell_arr,bam_arr = [],[]
        fetch_arr = [self.inbam] * len(self.cell_dict)
        strand_arr = [self.strand_dict] * len(self.cell_dict)
        qual_arr = [self.qual] * len(self.cell_dict)
        for x in self.cell_dict:
            tmpbamfile = f'{self.tmp_dir}/tmp_{x}.bam'
            bam_arr.append(tmpbamfile)
            cell_arr.append(self.cell_dict[x])
        self.bam_list = bam_arr

        mincpu = min(self.cell_num, self.thread)
        with Pool(mincpu) as pool:
            results = pool.starmap(Conversion.addTags,
                        zip(fetch_arr,bam_arr,cell_arr,strand_arr,qual_arr))
        return results
        

    @utils.add_log
    def snp_candidate(self, df_arr):
        Outputdf = pd.DataFrame()
        for i in df_arr:
            Outputdf = pd.concat([Outputdf,i])  
        Outputdf = Outputdf.reset_index()
        # all conv sites
        self.conv_df = Outputdf.groupby('index').agg({'convs':'sum','cells':'sum'})
        self.conv_df = self.conv_df.reset_index()
        self.conv_df[['chrom', 'pos']] = self.conv_df['index'].str.split('+', expand=True)
        self.conv_df = self.conv_df.set_index(['chrom', 'pos'])
        self.conv_df.drop('index', axis=1, inplace=True)
        # snp sites
        if self.snp_min_cells < 1:
            self.snp_min_cells = int(self.snp_min_cells * self.cell_num)
        self.snp_df = self.conv_df[ self.conv_df['cells'] >= self.snp_min_cells ]
        self.snp_df = self.snp_df[ self.snp_df['convs'] >= self.snp_min_depth ]
        # output
        self.conv_df.to_csv(self.outfile_csv)
        self.snp_df.to_csv(self.outsnp_csv)

    @utils.add_log
    def output_bam(self):
        if len(self.bam_list) > 1:
            bam_list = " ".join(self.bam_list)
            cmd = [f'samtools merge -@ {self.thread} -o {self.outtmp_bam}', f'{bam_list}']
            self.run_cmd(cmd)
        else:
            self.outtmp_bam = self.bam_list[0]
        cmd = [f'samtools sort -@ {self.thread} -o {self.outfile_bam}',f'{self.outtmp_bam}']
        self.run_cmd(cmd)
        cmd = ['rm', f'{self.outtmp_bam}']
        self.run_cmd(cmd)       

    @utils.add_log
    def clean_tmp(self):
        cmd = (f'rm -rf {self.tmp_dir}')
        self.debug_subprocess_call(cmd)

    @staticmethod
    def createTag(d):
        return ''.join([''.join(key) + str(d[key]) + ';' for key in d.keys()])[:-1]

    @staticmethod
    def convInRead(read, qual=20):
        specific_conversions = {}
        total_content = {'a': 0, 'c': 0, 'g': 0, 't': 0}
        specific_conversions[('c', 'A')] = 0
        specific_conversions[('g', 'A')] = 0
        specific_conversions[('t', 'A')] = 0
        specific_conversions[('a', 'C')] = 0
        specific_conversions[('g', 'C')] = 0
        specific_conversions[('t', 'C')] = 0
        specific_conversions[('a', 'G')] = 0
        specific_conversions[('c', 'G')] = 0
        specific_conversions[('t', 'G')] = 0
        specific_conversions[('a', 'T')] = 0
        specific_conversions[('c', 'T')] = 0
        specific_conversions[('g', 'T')] = 0
        specific_conversions[('a', 'N')] = 0
        specific_conversions[('c', 'N')] = 0
        specific_conversions[('g', 'N')] = 0
        specific_conversions[('t', 'N')] = 0

        tC_loc = []
        aG_loc = []

        try:
            refseq = read.get_reference_sequence().lower()
        except (UnicodeDecodeError):
            return 0
        except (AssertionError):
            return 0

        for base in total_content.keys():
            total_content[base] += refseq.count(base)
        for pair in read.get_aligned_pairs(with_seq=True):
            try:
                if pair[0] is not None and pair[1] is not None and pair[2] is not None:
                    if str(pair[2]).islower() and not read.query_qualities[pair[0]] < qual:
                        specific_conversions[(pair[2], read.seq[pair[0]])] += 1
                        if (pair[2], read.seq[pair[0]]) == ('t', 'C'):
                            tC_loc.append(pair[1])
                        if (pair[2], read.seq[pair[0]]) == ('a', 'G'):
                            aG_loc.append(pair[1])
            except (UnicodeDecodeError, KeyError):
                continue
        SC_tag = Conversion.createTag(specific_conversions)
        TC_tag = Conversion.createTag(total_content)

        if len(tC_loc) == 0:
            tC_loc.append(0)
        if len(aG_loc) == 0:
            aG_loc.append(0)
        return SC_tag, TC_tag, tC_loc, aG_loc

    @staticmethod
    def addTags(bamfilename, tmpoutbam, cell_list, strandedness, qual=20):
        tmp_cell = defaultdict(set)
        site_cell = defaultdict(int)
        site_depth = defaultdict(int)
        save = pysam.set_verbosity(0)
        bamfile = pysam.AlignmentFile(bamfilename, 'rb')
        header = bamfile.header
        mod_bamfile = pysam.AlignmentFile(tmpoutbam, mode='wb', header=header,check_sq=False)
        pysam.set_verbosity(save)

        class GeneError(Exception):
            pass

        for read in bamfile.fetch(until_eof=True):
            try:
                ## check read info
                if (not read.has_tag('GX')) or read.get_tag('GX') == '-':
                    continue
                if read.get_tag("CB") not in cell_list:
                    continue
                if read.get_tag('GX') not in strandedness:
                    raise GeneError

                tags = Conversion.convInRead(read, qual)
                if tags==0: 
                    continue
                read.set_tag('SC', tags[0], 'Z')
                read.set_tag('TC', tags[1], 'Z')
                read.set_tag('TL', tags[2])
                read.set_tag('AL', tags[3])
                read.set_tag('ST', strandedness[read.get_tag('GX')])
                mod_bamfile.write(read)

                if strandedness[read.get_tag('GX')] == '+':
                    locs = tags[2]
                else:
                    locs = tags[3]
                if locs[0] != 0:
                    for loc in locs:
                        site = f'{read.reference_name}+{loc}'
                        tmp_cell[site].add(read.get_tag("CB"))
                        site_depth[site] += 1
                  
            except (ValueError, KeyError):
                continue
            except (GeneError):
                print('{} is not in gtf file, please check your files.'.format(read.get_tag("GX")))
                continue
            #except (Exception):
            #    print("convert error")
            #    sys.exit(1)
        bamfile.close()
        mod_bamfile.close()

        for x in tmp_cell:
            site_cell[x] = len(tmp_cell[x])
        df1 = pd.DataFrame.from_dict(site_depth, orient='index')
        df2 = pd.DataFrame.from_dict(site_cell, orient='index')
        df = pd.concat([df1,df2],axis=1)
        df.columns=['convs','cells']
        return df
        


    @utils.add_log
    def add_conversion_metrics(self):
        self.add_metric(
            name='Conversion sites',
            value=self.conv_df.shape[0],
            help_info='the number of T_to_C conversion sites'
        )
        self.add_metric(
            name='Possible SNP sites',
            value=self.snp_df.shape[0],
            help_info='the number of possible SNP sites according to the conditions'
        )

    @utils.add_log
    def conversion_plot(self):
        self.conv_df['cell_pct'] = self.conv_df['cells'] / self.cell_num
        subplot = Conversion_plot(df_bar=self.conv_df).get_plotly_div()
        self.add_data(conversion_box=subplot)



@utils.add_log
def conversion(args):

    with Conversion(args) as runner:
        runner.run()


def get_opts_conversion(parser, sub_program):
    parser.add_argument('--genomeDir', help=HELP_DICT['genomeDir'])
    parser.add_argument('--basequalilty', default=20, type=int,
                        help='min base quality of the read sequence', required=False)
    parser.add_argument('--snp_min_cells', default=10, type=float,
                        help='Minimum number of cells to call a variant(>=1 for cell number or <1 for cell fraction), default 10.')
    parser.add_argument('--snp_min_depth',default=20, type=int,
                        help='Minimum depth to call a variant')
    parser.add_argument("--cellsplit", default=300, type=int, help='split N cells into a list')
    parser.add_argument('--conversionMem', default=30, type=int,
                        help='Default `30`. Set conversion memory.')

    if sub_program:
        parser.add_argument("--bam", 
            help='featureCount bam(sortedByCoord), must have "MD" tag, set in star step', required=True)
        parser.add_argument("--cell", help='barcode cell list', required=True)
        parser = s_common(parser)
    return parser





#!/bin/env python
# coding=utf8

import os
import pysam
import re
import pandas as pd
from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.tools.plotly_plot import Substitution_plot



class Substitution(Step):
    """
    ## Features
    - Computes the overall conversion rates in reads and plots a barplot.

    ## Output
    - `{sample}.substitution.txt` Tab-separated table of the overall conversion rates.
    """

    def __init__(self, args):
        Step.__init__(self, args)

        # input files
        self.sample = args.sample
        self.bam_file = args.bam
        self.outdir = args.outdir

        # output files
        self.outstat = os.path.join(self.outdir, self.sample+'.substitution.txt')

    @utils.add_log
    def run(self):
        # overall rate
        for_base, rev_base, is_forward, is_reverse = self.get_sub_tag(self.bam_file)
        self.sub_stat(for_base, rev_base, is_forward, is_reverse, self.outstat)
        #self.add_substitution_metrics()
        self.substitution_plot()
        self.add_help()

    @utils.add_log
    def get_sub_tag(self, bam):
        save = pysam.set_verbosity(0)
        bamfile = pysam.AlignmentFile(bam, 'rb', require_index=False)
        pysam.set_verbosity(save)
        is_reverse = {'cA': 0, 'gA': 0, 'tA': 0, 'aC': 0, 'gC': 0,
                      'tC': 0, 'aG': 0, 'cG': 0, 'tG': 0, 'aT': 0, 'cT': 0, 'gT': 0}
        is_forward = {'cA': 0, 'gA': 0, 'tA': 0, 'aC': 0, 'gC': 0,
                      'tC': 0, 'aG': 0, 'cG': 0, 'tG': 0, 'aT': 0, 'cT': 0, 'gT': 0}
        for_base = {'a': 0, 'c': 0, 'g': 0, 't': 0}
        rev_base = {'a': 0, 'c': 0, 'g': 0, 't': 0}
        snp_tags = ['', 'cA', 'gA', 'tA', 'aC', 'gC', 'tC', 'aG', 'cG', 'tG', 'aT', 'cT', 'gT']
        ref_tags = ['', 'a', 'c', 'g', 't']
        
        for read in bamfile.fetch(until_eof=True):
            try:
                snpmatch = re.match(
                    r'cA(\d+);gA(\d+);tA(\d+);aC(\d+);gC(\d+);tC(\d+);aG(\d+);cG(\d+);tG(\d+);aT(\d+);cT(\d+);gT(\d+);', read.get_tag('SC'), re.M)
                totmatch = re.match(r'a(\d+);c(\d+);g(\d+);t(\d+)', read.get_tag('TC'), re.M)
                if snpmatch and totmatch:
                    if read.is_reverse:
                        for j in range(1, len(ref_tags)):
                            rev_base[ref_tags[j]] += int(totmatch.group(j))
                        for i in range(1, len(snp_tags)):
                            is_reverse[snp_tags[i]] += int(snpmatch.group(i))
                    else:
                        for j in range(1, len(ref_tags)):
                            for_base[ref_tags[j]] += int(totmatch.group(j))
                        for i in range(1, len(snp_tags)):
                            is_forward[snp_tags[i]] += int(snpmatch.group(i))
            except (ValueError, KeyError):
                continue
        bamfile.close()

        return for_base, rev_base, is_forward, is_reverse

    @utils.add_log
    def sub_stat(self, for_base, rev_base, is_forward, is_reverse, outfile):
        convertdict = {'a': ['aC', 'aG', 'aT'],
                       'c': ['cA', 'cG', 'cT'],
                       'g': ['gA', 'gC', 'gT'],
                       't': ['tA', 'tC', 'tG']}
        subdict = {'a': 't', 't': 'a', 'c': 'g', 'g': 'c',
                   'aC': 'tG', 'aG': 'tC', 'aT': 'tA',
                   'cA': 'gT', 'cG': 'gC', 'cT': 'gA',
                   'gA': 'cT', 'gC': 'cG', 'gT': 'cA',
                   'tA': 'aT', 'tC': 'aG', 'tG': 'aC'}
        outdict = {'aC': 'A_to_C', 'aG': 'A_to_G', 'aT': 'A_to_T',
                   'cA': 'C_to_A', 'cG': 'C_to_G', 'cT': 'C_to_T',
                   'gA': 'G_to_A', 'gC': 'G_to_C', 'gT': 'G_to_T',
                   'tA': 'T_to_A', 'tC': 'T_to_C', 'tG': 'T_to_G'}
        outw = open(outfile, 'w')
        for x in ['a', 'c', 'g', 't']:
            fbase = for_base[x]
            rbase = rev_base[subdict[x]]
            for y in convertdict[x]:
                fcov = is_forward[y]*100 / float(fbase)
                rcov = is_reverse[subdict[y]]*100 / float(rbase)
                outw.write(outdict[y]+'\t'+"%.3f" % fcov+'\t'+"%.3f" % rcov+'\n')
        outw.close()

    @utils.add_log
    def substitution_plot(self):
        df = pd.read_table(self.outstat, header=None)
        df.columns = ['sample', '+', '-']
        subplot = Substitution_plot(df_bar=df).get_plotly_div()
        self.add_data(substitution_box=subplot)

    @utils.add_log
    def add_help(self):
        self.add_help_content(
            name='Substitution:',
            content='Nucleotide substitution rate per conversion type'
        )


@utils.add_log
def substitution(args):

    with Substitution(args) as runner:
        runner.run()


def get_opts_substitution(parser, sub_program):
    if sub_program:
        parser.add_argument('--bam', help='bam file from conversion step', required=True)
        parser = s_common(parser)
    return parser



import argparse
import glob
import itertools
import os
import sys
from collections import defaultdict

import celescope
from celescope.tools.__init__ import FILTERED_MATRIX_DIR_SUFFIX, STAR_BAM_SUFFIX
from celescope.tools import utils
from celescope.celescope import ArgFormatter
from celescope.__init__ import HELP_DICT
from celescope.tools.make_ref import MakeRef

TOOLS_DIR = os.path.dirname(celescope.tools.__file__)


class Multi():

    def __init__(self, assay):
        self.__ASSAY__ = assay
        init_module = utils.find_assay_init(assay)
        self.STEPS = init_module.STEPS
        self.REMOVE_FROM_MULTI = getattr(init_module, 'REMOVE_FROM_MULTI', set())
        self.REMOVE_FROM_MULTI.add('mkref')
        try:
            self.__CONDA__ = os.path.basename(os.environ['CONDA_DEFAULT_ENV'])
        except KeyError:
            print('CONDA_DEFAULT_ENV is not set. sjm mode may not available.')
            self.__CONDA__ = 'celescope'
        self.__APP__ = 'celescope'

        # remove
        for step in self.REMOVE_FROM_MULTI:
            if step in self.STEPS:
                self.STEPS.remove(step)

        # add args
        self.parser = None
        self.common_args()
        self.step_args()

        # set
        self.args = None
        self.col4_default = None
        self.last_step = ''
        self.fq_suffix = ""
        self.steps_run = self.STEPS
        self.fq_dict = {}
        self.col4_dict = {}
        self.col5_dict = {}
        self.logdir = None
        self.sjm_dir = None
        self.sjm_file = None

        self.sjm_cmd = ''
        self.sjm_order = ''
        self.shell_dict = defaultdict(str)

        self.outdir_dic = {}

    def common_args(self):
        readme = f'{self.__ASSAY__} multi-samples'
        parser = argparse.ArgumentParser(readme,
                                         formatter_class=ArgFormatter,
                                         conflict_handler='resolve')
        parser.add_argument(
            '--mapfile',
            help='''
Mapfile is a tab-delimited text file with as least three columns. Each line of mapfile represents paired-end fastq files.

1st column: Fastq file prefix.  
2nd column: Fastq file directory path.  
3rd column: Sample name, which is the prefix of all output files.  
4th column: The 4th column has different meaning for each assay. The single cell rna directory after running CeleScope is called `matched_dir`.

- `vdj` Required, matched_dir.
- `tag` Required, matched_dir.
- `dynaseq` Optional, forced cell number.
- `snp` Required, matched_dir.
- `capture_virus` Required, matched_dir.
- `fusion` Required, matched_dir.
- `citeseq` Required, matched_dir.
- `flv_trust4` Required, matched_dir.
- `sweetseq` Required, matched_dir.
 
5th column:
- `dynaseq` Required, background snp file.

Example

Sample1 has 2 paired-end fastq files located in 2 different directories(fastq_dir1 and fastq_dir2). Sample2 has 1 paired-end fastq file located in fastq_dir1.
```
$cat ./my.mapfile
fastq_prefix1	fastq_dir1	sample1
fastq_prefix2	fastq_dir2	sample1
fastq_prefix3	fastq_dir1	sample2

$ls fastq_dir1
fastq_prefix1_1.fq.gz	fastq_prefix1_2.fq.gz
fastq_prefix3_1.fq.gz	fastq_prefix3_2.fq.gz

$ls fastq_dir2
fastq_prefix2_1.fq.gz	fastq_prefix2_2.fq.gz
```
''',
            required=True)
        parser.add_argument('--mod', help='Which type of script to generate, `sjm` or `shell`.',
            choices=['sjm', 'shell'], default='sjm')
        parser.add_argument('--queue', help='Only works if the `--mod` selects `sjm`.')
        parser.add_argument('--rm_files', action='store_true',
            help='Remove redundant fastq and bam files after running.')
        parser.add_argument('--steps_run', 
            help='''
Steps to run. Multiple Steps are separated by comma. For example, if you only want to run `barcode` and `cutadapt`, 
use `--steps_run barcode,cutadapt`
''', 
            default='all')
        # sub_program parser do not have
        parser.add_argument('--outdir', help='Output directory.', default="./")
        parser.add_argument('--thread', help=HELP_DICT['thread'], default=4)
        parser.add_argument('--use_R3', help='ATAC libraries use R3 reads instead of R2.', action='store_true')
        parser.add_argument('--debug', help=HELP_DICT['debug'], action='store_true')
        self.parser = parser
        return parser

    def step_args(self):
        for step in self.STEPS:
            step_module = utils.find_step_module(self.__ASSAY__, step)
            func_opts = getattr(step_module, f"get_opts_{step}")
            func_opts(self.parser, sub_program=False)

    @staticmethod
    @utils.add_log
    def parse_mapfile(mapfile, default_val, use_R3=False):
        fq_dict = defaultdict(list)
        col4_dict = {}
        col5_dict = {}
        with open(mapfile) as fh:
            for line in fh:
                line = line.strip()
                if not line or line.startswith('#'):
                    continue
                line_split = line.split()
                library_id, library_path, sample_name = line_split[:3]
                if len(line_split) >= 4:
                    col4 = line_split[3]
                else:
                    col4 = default_val
                fq1, fq2 = get_fq(library_id, library_path, use_R3)

                if sample_name in fq_dict:
                    fq_dict[sample_name][0].append(fq1)
                    fq_dict[sample_name][1].append(fq2)
                else:
                    fq_dict[sample_name] = [[fq1], [fq2]]
                    col4_dict[sample_name] = col4
                if len(line_split) == 5:
                    col5_dict[sample_name] = line_split[4]

        for sample_name in fq_dict:
            fq_dict[sample_name][0] = ",".join(fq_dict[sample_name][0])
            fq_dict[sample_name][1] = ",".join(fq_dict[sample_name][1])

        if not fq_dict:
            raise Exception('empty mapfile!')
        return fq_dict, col4_dict, col5_dict

    def link_data(self):
        raw_dir = f'{self.args.outdir}/data_give/rawdata'
        os.system('mkdir -p %s' % (raw_dir))
        with open(raw_dir + '/ln.sh', 'w') as fh:
            fh.write('cd %s\n' % (raw_dir))
            for s, arr in self.fq_dict.items():
                fh.write('ln -sf %s %s\n' % (arr[0], s + '_1.fq.gz'))
                fh.write('ln -sf %s %s\n' % (arr[1], s + '_2.fq.gz'))

    def prepare(self):
        """
        parse_mapfile, make log dir, init script variables, init outdir_dic
        make sjm dir, sjm file
        """
        self.args = self.parser.parse_args()
        self.check_genome()

        if self.args.steps_run != 'all':
            self.steps_run = self.args.steps_run.strip().split(',')
        
        if self.args.mod == 'sjm':

            self.sjm_dir = f'{self.args.outdir}/sjm/'
            utils.check_mkdir(self.sjm_dir)
            self.logdir = self.args.outdir + '/log'
            utils.check_mkdir(self.logdir)

            self.sjm_file = f'{self.sjm_dir}/sjm.job'
            self.sjm_cmd = f'log_dir {self.logdir}\n'

        # parse_mapfile
        self.fq_dict, self.col4_dict, self.col5_dict = self.parse_mapfile(self.args.mapfile, self.col4_default, self.args.use_R3)

        for sample in self.fq_dict:
            self.outdir_dic[sample] = {}
            index = 0
            for step in self.STEPS:
                step_outdir = f"{self.args.outdir}/{sample}/{index:02d}.{step}"
                self.outdir_dic[sample].update({step: step_outdir})
                index += 1
            # add outs dir
            self.outdir_dic[sample]['outs'] = f"{self.args.outdir}/{sample}/outs"

    def generate_cmd(self, cmd, step, sample, m=1, x=1):
        if sample:
            sample = "_" + sample
        sched_options = f'sched_options -w n -cwd -V -l vf={m}g,p={x}'
        if self.args.queue:
            sched_options += f' -q {self.args.queue} '
        self.sjm_cmd += f'''
job_begin
    name {step}{sample}
    {sched_options}
    cmd source activate {self.__CONDA__}; {cmd}
job_end
'''

    def process_cmd(self, cmd, step, sample, m=1, x=1):
        self.generate_cmd(cmd, step, sample, m=m, x=x)
        self.shell_dict[sample] += cmd + '\n'
        if self.last_step:
            self.sjm_order += f'order {step}_{sample} after {self.last_step}_{sample}\n'
        self.last_step = step

    def parse_step_args(self, step):
        step_module = utils.find_step_module(self.__ASSAY__, step)
        func_opts = getattr(step_module, f"get_opts_{step}")
        step_parser = argparse.ArgumentParser()
        func_opts(step_parser, sub_program=False)
        args = step_parser.parse_known_args()
        return args

    def get_cmd_line(self, step, sample):
        """ get cmd line without input
            return str
        """
        args = self.parse_step_args(step)
        args_dict = args[0].__dict__
        step_prefix = (
            f'{self.__APP__} {self.__ASSAY__} {step} '
            f'--outdir {self.outdir_dic[sample][step]} '
            f'--sample {sample} '
            f'--thread {self.args.thread} '
        )
        cmd_line = step_prefix
        if self.args.debug:
            cmd_line += " --debug "
        for arg in args_dict:
            if args_dict[arg] is False:
                continue
            if args_dict[arg] is True:
                cmd_line += f'--{arg} '
            else:
                if args_dict[arg]:
                    matches = [' ', '-']
                    arg_string = str(args_dict[arg])
                    if any(char in arg_string for char in matches):  # need quote
                        cmd_line += f'--{arg} "{arg_string}" '
                    else:
                        cmd_line += f'--{arg} {arg_string} '

        return cmd_line

    def sample(self, sample):
        step = "sample"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def barcode(self, sample):
        step = "barcode"
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def cutadapt(self, sample):
        step = "cutadapt"
        fq = f'{self.outdir_dic[sample]["barcode"]}/{sample}_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)

    def star(self, sample):
        step = 'star'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)


    def prep_map(self, sample):
        step = 'prep_map'
        arr = self.fq_dict[sample]
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq1 {arr[0]} --fq2 {arr[1]} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def featureCounts(self, sample):
        step = 'featureCounts'
        if "star" in self.STEPS:
            prev = 'star'
        elif "prep_map" in self.STEPS:
            prev = 'prep_map'
        else:
            sys.exit('To use featureCounts, star or prep must in the steps!')
        input_bam = f'{self.outdir_dic[sample][prev]}/{sample}_{STAR_BAM_SUFFIX}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--input {input_bam} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=self.args.thread)

    def count(self, sample):
        step = 'count'
        count_detail = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_count_detail.txt'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--count_detail {count_detail} '
            f'--force_cell_num {self.col4_dict[sample]} '
        )

        self.process_cmd(cmd, step, sample, m=20, x=1)

    def analysis(self, sample):
        step = 'analysis'
        matrix_file = f'{self.outdir_dic[sample]["count"]}/{sample}_{FILTERED_MATRIX_DIR_SUFFIX}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--matrix_file {matrix_file} '
        )
        self.process_cmd(cmd, step, sample, m=10, x=1)

    def consensus(self, sample):
        step = 'consensus'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)
        outfile = f'{self.outdir_dic[sample][step]}/{sample}_consensus.fq'
        return outfile

    def run_steps(self):
        for sample in self.fq_dict:
            self.last_step = ''
            for step in self.steps_run:
                try:
                    method_to_call = getattr(self, step)
                except AttributeError as attr_not_exist:
                    raise NotImplementedError(
                        "Class `{}` does not implement `{}`".format(self.__class__.__name__, step)
                    ) from attr_not_exist
                method_to_call(sample)

    def merge_report(self):
        step = "merge_report"
        steps_str = ",".join(self.STEPS)
        steps_str = steps_str.replace('prep_map', 'barcode,cutadapt,star')
        steps_str = steps_str.replace('starsolo', 'demultiplexing,mapping,cells')
        samples = ','.join(self.fq_dict.keys())
        app = TOOLS_DIR + '/merge_table.py'
        cmd = (
            f'python {app} --samples {samples} '
            f'--steps {steps_str} --outdir {self.args.outdir}'
        )
        if self.args.rm_files:
            cmd += ' --rm_files'
        self.generate_cmd(cmd, step, sample="")
        for sample in self.fq_dict:
            self.sjm_order += f'order {step} after {self.last_step}_{sample}\n'

    def end(self):
        if self.args.mod == 'sjm':
            self.merge_report()
            with open(self.sjm_file, 'w') as fh:
                fh.write(self.sjm_cmd + '\n')
                fh.write(self.sjm_order)
        if self.args.mod == 'shell':
            os.system('mkdir -p ./shell/')
            for sample in self.shell_dict:
                with open(f'./shell/{sample}.sh', 'w') as f:
                    f.write(self.shell_dict[sample])

    def check_genome(self):
        for arg, val in vars(self.args).items():
            if val and 'genomeDir' in arg:
                MakeRef.get_config(val)


    def run(self):
        self.prepare()
        self.run_steps()
        self.end()


def get_read(library_id, library_path, read='1'):
    read1_list = [f'_{read}', f'R{read}', f'R{read}_001']
    fq_list = ['fq', 'fastq']
    suffix_list = ["", ".gz"]
    read_pattern_list = [
        f'{library_path}/{library_id}*{read}.{fq_str}{suffix}'
        for read in read1_list
        for fq_str in fq_list
        for suffix in suffix_list
    ]
    fq_list = [glob.glob(read1_pattern) for read1_pattern in read_pattern_list]
    fq_list = (non_empty for non_empty in fq_list if non_empty)
    fq_list = sorted(list(itertools.chain(*fq_list)))
    if len(fq_list) == 0:
        print("Allowed R1 patterns:")
        for pattern in read_pattern_list:
            print(pattern)
        raise Exception(
            '\n'
            f'Invalid Read{read} path! \n'
            f'library_id: {library_id}\n'
            f'library_path: {library_path}\n'
        )
    return fq_list


def get_fq(library_id, library_path, use_R3=False):
    fq1_list = get_read(library_id, library_path, read='1')
    if use_R3:
        fq2_list = get_read(library_id, library_path, read='3')
    else:
        fq2_list = get_read(library_id, library_path, read='2')
    if len(fq1_list) != len(fq2_list):
        raise Exception("Read1 and Read2 fastq number do not match!")
    fq1 = ",".join(fq1_list)
    fq2 = ",".join(fq2_list)
    return fq1, fq2



import pandas as pd

from celescope.tools.step import Step, s_common
from celescope.tools.plotly_plot import Tsne_dropdown_plot,Tsne_single_plot,Tsne_plot



def get_opts_analysis_cite(parser, sub_program):
    if sub_program:
        parser.add_argument('--tsne_coord', help='tsne coord file', required=True)
        s_common(parser)


def analysis_cite(args):
    with Analysis_cite(args, display_title='Analysis') as runner:
        runner.run()


class Analysis_cite(Step):
    def __init__(self, args, display_title):
        super().__init__(args, display_title)

        # input_file
        self.tsne_coord = self.args.tsne_coord

    def run(self):
        df_tsne = pd.read_csv(self.tsne_coord,sep="\t",index_col=0)
        feature_name_list = [x[:45] for x in df_tsne.columns[4:]]
        df_tsne.columns = df_tsne.columns[:4].to_list()+feature_name_list

        # accelerate
        df_tsne['tSNE_1'] = df_tsne['tSNE_1'].astype('float16')
        df_tsne['tSNE_2'] = df_tsne['tSNE_2'].astype('float16')
        for col in feature_name_list:
            df_tsne[col] = df_tsne[col].astype('float16')

        # plot
        tsne_cluster = Tsne_plot(df_tsne, 'cluster').get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)
        tsne_citeseq = Tsne_dropdown_plot(df_tsne,'Citeseq',feature_name_list).get_plotly_div()
        self.add_data(tsne_citeseq=tsne_citeseq)
        Tsne_single_plot(df_tsne,feature_name_list,self.args.outdir).get_plotly_div()



import pandas as pd
import numpy as np

from celescope.tools.plotly_plot import Tsne_plot
from celescope.tools.step import Step
from celescope.tools.capture.__init__ import SUM_UMI_COLNAME, LOG2_COLNAME
from celescope.tools import analysis_wrapper

def get_opts_analysis(parser, sub_program):
    if sub_program:
        parser.add_argument('--filter_umi_file', help='filter umi file', required=True)
        analysis_wrapper.get_opts_analysis_match(parser, sub_program)


class Analysis(Step):

    def __init__(self, args, display_title='Analysis'):
        super().__init__(args, display_title)
        self.df_filter_umi = pd.read_csv(args.filter_umi_file, index_col=0)

        # out
        self.df_tsne_file = f'{self.out_prefix}_UMI_tsne.csv'
        self.df_tsne = None

    def add_tsne_info(self):
        """
        add tsne coord and log2 column
        """
        report_runner = analysis_wrapper.Report_runner(self.args)
        df_tsne, _df_marker = report_runner.get_df()
        self.df_tsne = pd.merge(df_tsne, self.df_filter_umi, left_index=True, right_index=True)
        self.df_tsne[LOG2_COLNAME] = np.log2(self.df_tsne[SUM_UMI_COLNAME] + 1)
        self.df_tsne.to_csv(self.df_tsne_file)

    def add_cluster_metrics(self):
        self.add_help_content('cluster 1,2,3...', 'number of positive cells(sum_UMI > 0) in each cluster after filtering')

        df_cluster_all = self.df_tsne.groupby("cluster").count()

        df_positive = self.df_tsne[self.df_tsne[SUM_UMI_COLNAME] > 0]
        df_cluster_positive = df_positive.groupby("cluster").count()

        for index, row in df_cluster_positive.iterrows():
            self.add_metric(
                name=f'cluster {index}',
                value=int(row['tSNE_1']),
                total=int(df_cluster_all.loc[index, 'tSNE_1']),
            )

    def run(self):
        self.add_tsne_info()
        self.add_cluster_metrics()

        tsne_cluster = Tsne_plot(self.df_tsne, 'cluster').get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)

        tsne_plot = Tsne_plot(self.df_tsne, LOG2_COLNAME, discrete=False)
        tsne_plot.set_color_scale(['LightGrey', 'Orange', 'Red'])
        tsne_feature = tsne_plot.get_plotly_div()
        self.add_data(tsne_feature=tsne_feature)


import os
from collections import OrderedDict

__VERSION__ = "2.0.2"
__version__ = __VERSION__

ASSAY_LIST = [
    "rna",
    'vdj',
    'tag',
    'dynaseq',
    'snp',
    'capture_virus',
    'fusion',
    'citeseq',
    'flv_trust4',
    'sweetseq',
    'utils',
    'bulk_vdj',
    'bulk_rna',
]

ROOT_PATH = os.path.dirname(__file__)

RELEASED_ASSAYS = ['rna', 'vdj', 'tag', 'dynaseq', 'snp', 'capture_virus', 'fusion', 'citeseq', 'flv_trust4', 'sweetseq','bulk_vdj','bulk_rna']

# argument help
HELP_DICT = {
    'match_dir': 'Match celescope scRNA-Seq directory.',
    'gene_list': 'Required. Gene list file, one gene symbol per line. Only results of these genes are reported. Conflict with `--panel`',
    'genomeDir': 'Genome directory after running `celescope {assay} mkref`.',
    'thread': 'Thread to use.',
    'debug': 'If this argument is used, celescope may output addtional file for debugging.',
    'fasta': 'Required. Genome fasta file. Use absolute path or relative path to `genomeDir`.',
    'outdir': 'Output directory.',
    'matrix_dir': 'Match celescope scRNA-Seq matrix directory.',
    'panel': 'The prefix of bed file in `celescope/data/snp/panel/`, such as `lung_1`. Conflict with `--gene_list`',
    'virus_genomeDir': 'Required. Virus genome directory after running `celescope capture_virus mkref`.',
    'threshold_method': 'One of [otsu, auto, hard, none].',
    'tsne_file': 'match_dir t-SNE coord file. Do not required when `--match_dir` is provided.',
    'df_marker_file': 'match_dir df_marker_file. Not required when `--match_dir` is provided.',
    'cell_calling_method': 'Default `EmptyDrops_CR`. Choose from [`auto`, `EmptyDrops_CR`]',
    'additional_param': 'Additional parameters for the called software. Need to be enclosed in quotation marks. For example, `--{software}_param "--param1 value1 --param2 value2"`.',
    'genomeSAindexNbases': '''For small genomes, the parameter --genomeSAindexNbases must to be scaled down, with a typical 
value of min(14, log2(GenomeLength)/2 - 1). For example, for 1 megaBase genome, this is equal 
to 9, for 100 kiloBase genome, this is equal to 7.''',
    'chemistry': '`--chemistry auto` can auto-detect scopeV2 mRNA, scopeV3 mRNA, full length VDJ mRNA(flv_rna) and full length VDJ(flv). You need to explicitly use `--chemistry scopeV1` for legacy chemistry scopeV1. `--chemistry customized` is used for user defined combinations that you need to provide `--pattern`, `--whitelist` and `--linker` at the same time.',   
}

# report metrics help
HELP_INFO_DICT = {
    'matched_barcode_number': {
        'display': 'Number of Matched Cells',
        'info': 'cell barcode number of matched scRNA-Seq sample',
    }
}



import pandas as pd

from celescope.tools import utils
from celescope.tools.step import Step
from celescope.tools.step import s_common
from celescope.tools.plotly_plot import Tsne_plot
from celescope.tools import analysis_wrapper
from celescope.__init__ import HELP_DICT


class Analysis_tag(Step):
    """
    ## Features
    - Combine scRNA-Seq clustering infromation with tag assignment.
    """

    def __init__(self, args, display_title=None):
        super().__init__(args, display_title)
        self.tsne_tag_file = args.tsne_tag_file
        self.df_tsne_tag = pd.read_csv(self.tsne_tag_file, sep='\t')

    def run(self):
        report_runner = analysis_wrapper.Report_runner(self.args, display_title=self.display_title)
        df_tsne, _df_marker = report_runner.get_df()

        tsne_cluster = Tsne_plot(df_tsne, 'cluster').get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)

        tsne_tag = Tsne_plot(self.df_tsne_tag, 'tag').get_plotly_div()
        self.add_data(tsne_tag=tsne_tag)


def get_opts_analysis_tag(parser, sub_program):
    if sub_program:
        parser.add_argument('--tsne_tag_file', help='`{sample}_tsne_tag.tsv` from count_tag. ', required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--tsne_file", help=HELP_DICT['tsne_file'])
        parser = s_common(parser)


@utils.add_log
def analysis_tag(args):
    with Analysis_tag(args, display_title="Analysis") as runner:
        runner.run()



from celescope.fusion.__init__ import __ASSAY__
from celescope.tools.multi import Multi
from celescope.tools.__init__ import STAR_BAM_SUFFIX

class Multi_fusion(Multi):
    """

    ## Usage
    ```
    multi_fusion\\
    --mapfile ./fusion.mapfile\\
    --fusion_genomeDir {fusion_genomeDir}\\  
    --mod shell\\
    ```

    Use `celescope/data/fusion/{panel}/mkref.sh` to generate the fusion genomeDir. 
    ```
    source activate celescope
    celescope fusion mkref \\
    --genome_name BCR_ABL1_PML_RARA_all \\
    --fasta BCR_ABL1_PML_RARA_all.fasta \\
    --fusion_pos BCR_ABL1_PML_RARA_all_pos.txt\\
    ```

    ## Main Output
    - `05.filter_fusion/{sample}_filtered_UMI.csv`: Filtered fusion UMI counts of each cell barcode.
    """

    def star_fusion(self, sample):
        step = 'star_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def count_fusion(self, sample):
        step = 'count_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        capture_bam = f'{self.outdir_dic[sample]["star_fusion"]}/{sample}_{STAR_BAM_SUFFIX}'
        cmd = (
            f'{cmd_line} '
            f'--capture_bam {capture_bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def filter_fusion(self, sample):
        step = 'filter_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        raw_read_count_file = f'{self.outdir_dic[sample]["count_fusion"]}/{sample}_raw_read_count.json'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--raw_read_count_file {raw_read_count_file} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def analysis_fusion(self, sample):
        step = 'analysis_fusion'
        cmd_line = self.get_cmd_line(step, sample)
        filter_umi_file = f'{self.outdir_dic[sample]["filter_fusion"]}/{sample}_filtered_UMI.csv'
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--filter_umi_file {filter_umi_file} '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)


def main():
    multi = Multi_fusion(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



from celescope.tools.multi import Multi
from celescope.bulk_vdj.__init__ import __ASSAY__


class Multi_bulk_vdj(Multi):
    """
    ## Reference
    - Human
    ```
    mkdir -p /genome/vdj/human
    cd /genome/vdj/human
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/TR/TR{A,B}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/TR/TRBD.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/IG/IG{H,K,L}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Homo_sapiens/IG/IGHD.fasta
    celescope vdj mkref human TR
    celescope vdj mkref human IG
    ```
    - Mouse
    ```
    mkdir -p /genome/vdj/mouse
    cd /genome/vdj/mouse
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/TR/TR{A,B}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/TR/TRBD.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/IG/IG{H,K,L}{V,J}.fasta
    wget http://www.imgt.org/download/V-QUEST/IMGT_V-QUEST_reference_directory/Mus_musculus/IG/IGHD.fasta
    celescope vdj mkref mouse TR
    celescope vdj mkref mouse IG
    ```
 
    ## Usage
    ```
    multi_bulk_vdj \\
        --mapfile ./vdj.mapfile \\
        --ref_path /genome/vdj/human/human_TR \\
        --species human \\
        --type TCR \\
        --min_consensus_read 2 \\
        --allowNoLinker \\
        --thread 8 \\
        --mod shell
    ``` 
    """
         
    def consensus(self, sample):
        step = 'consensus'
        fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
            f'--out_fasta '
        )
        self.process_cmd(cmd, step, sample, m=5, x=1)
        outfile = f'{self.outdir_dic[sample][step]}/{sample}_consensus.fasta'
        return outfile
    
    def mapping_vdj(self, sample):
        step = 'mapping_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        #fasta = f'{self.outdir_dic[sample]["consensus"]}/{sample}_consensus.fasta'
        fasta = f'{self.outdir_dic[sample]["consensus"]}/{sample}_filtered_consensus.fasta'
        cmd = (
            f'{cmd_line} '
            f'--fasta {fasta} '
        )
        self.process_cmd(cmd, step, sample, m=15, x=self.args.thread)

    def count_vdj(self, sample):
        step = 'count_vdj'
        cmd_line = self.get_cmd_line(step, sample)
        productive_file = f'{self.outdir_dic[sample]["mapping_vdj"]}/{sample}_productive.tsv'
        airr_file = f'{self.outdir_dic[sample]["mapping_vdj"]}/{sample}_airr.tsv'
        fq = f'{self.outdir_dic[sample]["barcode"]}/{sample}_2.fq{self.fq_suffix}'
        cmd = (
            f'{cmd_line} '
            f'--productive_file {productive_file} '
            f'--airr_file {airr_file} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=self.args.thread)


def main():
    multi = Multi_bulk_vdj(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()


import subprocess
import os

from celescope.tools import utils


class Mkref:
    """
    ## Features

    - Build Index for IMGT_ref.
    
    - Make sure current directory contains all V,D,J of TRA/TRB or IGH/IGK/IGL reference downloaded from IMGT website.

    ## Usage
    ```
    celescope vdj mkref species(human or mouse) imgt_type(TR or IG)
    ```

    ## Output

    - VDJ IMGT reference with index files.

    """


    def __init__(self, args):
        self.type = args.imgt_type
        self.species=  args.species
        self.outdir = f"{os.getcwd()}/{self.species}_{self.type}"


    def __call__(self):
        utils.check_mkdir(self.outdir)
        self.combine_seq()
        self.build_index()


    @utils.add_log
    def combine_seq(self):
        """
        Combine all V, all D and all J sequences, respectively, into separate files:
        """
        imgt_files = os.listdir()
        if self.type == "TR":
            imgt_files = [i for i in imgt_files if i[:2]=="TR"]
            assert len(imgt_files) == 5
            cmd_V = f"cat TRAV.fasta TRBV.fasta > {self.outdir}/TRV.fasta"
            cmd_D = f"cat TRBD.fasta > {self.outdir}/TRD.fasta" 
            cmd_J = f"cat TRAJ.fasta TRBJ.fasta > {self.outdir}/TRJ.fasta"
        elif self.type == "TRGD":
            imgt_files = [i for i in imgt_files if i[:2]=="TR"]
            assert len(imgt_files) == 5
            cmd_V = f"cat TRDV.fasta TRGV.fasta > {self.outdir}/TRV.fasta"
            cmd_D = f"cat TRDD.fasta > {self.outdir}/TRD.fasta" 
            cmd_J = f"cat TRDJ.fasta TRGJ.fasta > {self.outdir}/TRJ.fasta"            
        else:
            imgt_files = [i for i in imgt_files if i[:2]=="IG"]
            assert len(imgt_files) == 7
            cmd_V = f"cat IGHV.fasta IGKV.fasta IGLV.fasta > {self.outdir}/IGV.fasta"
            cmd_D = f"cat IGHD.fasta > {self.outdir}/IGD.fasta" 
            cmd_J = f"cat IGHJ.fasta IGKJ.fasta IGLJ.fasta > {self.outdir}/IGJ.fasta"
        
        for cmd in [cmd_V, cmd_D, cmd_J]:
            subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def build_index(self):
        """
        build imgt index for igblast
        """
        os.chdir(self.outdir)
        for file in os.listdir():
            out_file_name, _ = os.path.splitext(file)
            subprocess.check_call(f"edit_imgt_file.pl {file} > {out_file_name}.fa", shell=True)
            subprocess.check_call(f"makeblastdb -parse_seqids -dbtype nucl -in {out_file_name}.fa", shell=True)


def mkref(args):
    runner = Mkref(args)
    runner()


def get_opts_mkref(parser, sub_program=True):
    if sub_program:
        parser.add_argument(
            'species',
            help='human or mouse.'
        )
        parser.add_argument(
            'imgt_type',
            help='TR: build index for TR IMGT reference, IG: build index for IG IMGT reference')

    return parser


from multiprocessing import Pool
import subprocess
import math

from celescope.tools.step import Step, s_common
from celescope.tools import utils
from celescope.flv_trust4.__init__ import CHAIN, REF_DIR

CANDIDATE_FQ_SUFFIX = 'bcrtcr.fq'


class Mapping(Step):
    """
    ## Features

    - Extract candidate reads to assemble.

    ## Output
    - `02.mapping/{sample}_bcrtcr.fq` All candidate reads(mapped to any V(D)J genes) sequence.
    - `02.mapping/{sample}_bcrtcr_bc.fa` All candidate reads(mapped to any V(D)J genes) barcode.
    - `02.mapping/{sample}_bcrtcr_umi.fa` All candidate reads(mapped to any V(D)J genes) umi.
    - `02.mapping/{sample}_{chain}.fq` Candidate reads(mapped to {chain}) sequence. For BCR: IGH, IGK, IGL, For TCR: TRA, TRB.
    - `02.mapping/{sample}_{chain}_bc.fa` Candidate reads(mapped to {chain}) barcode. For BCR: IGH, IGK, IGL, For TCR: TRA, TRB.
    - `02.mapping/{sample}_{chain}_umi.fa` Candidate reads(mapped to {chain}) umi. For BCR: IGH, IGK, IGL, For TCR: TRA, TRB.
    """
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        self._ref = args.ref
        self._barcodeRange = args.barcodeRange
        self._umiRange = args.umiRange
        self._match_fq1 = args.match_fq1
        self._match_fq2 = args.match_fq2
        self._chains = CHAIN[args.seqtype]

        n_extract = len(self._chains) + 1
        self._single_thread = math.ceil(self.thread / n_extract)
        self._matched_reads = self.get_slot_key(
            slot='metrics',
            step_name='barcode',
            key='Valid Matched Reads',
        )

    
    @utils.add_log
    def extract_chain_reads(self):
        """
        bcrtcr.fq + {chains}.fq
        """
        cb_range = self._barcodeRange.split(' ')
        umi_range = self._umiRange.split(' ')

        map_index_prefix = ['bcrtcr'] + self._chains
        n_map = len(map_index_prefix)
        samples = [self.sample] * n_map
        map_ref = [self._ref] * n_map
        map_outdirs = [self.outdir] * n_map
        map_fq1 = [self._match_fq1] * n_map
        map_fq2 = [self._match_fq2] * n_map
        map_cb_range = [cb_range] * n_map
        map_umi_range = [umi_range] * n_map
        map_n_thread = [self._single_thread] * n_map

        with Pool(n_map) as pool:
            pool.starmap(Mapping.extract_candidate_reads, 
            zip(map_ref, map_index_prefix, map_outdirs, samples, map_fq1, map_fq2, map_cb_range, map_umi_range, map_n_thread))  

    @staticmethod
    @utils.add_log
    def extract_candidate_reads(ref, index_prefix, outdir, sample, fq1, fq2, barcodeRange, umiRange, single_thread):
        """
        helper function for extract reads map to index_prefix
        index_prefix can be 'bcrtcr' + chains
        """
        cmd = (
            f'fastq-extractor -t {single_thread} '
            f'-f {REF_DIR}/{ref}/{index_prefix}.fa '
            f'-o {outdir}/{sample}_{index_prefix} '
            f'--barcodeStart {barcodeRange[0]} '
            f'--barcodeEnd {barcodeRange[1]} '
            f'--umiStart {umiRange[0]} '
            f'--umiEnd {umiRange[1]} '
            f'-u {fq2} '
            f'--barcode {fq1} '
            f'--UMI {fq1} '
            '2>&1 '
            )
        Mapping.extract_candidate_reads.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)  

    def add_metrics(self):
        n_bcrtcr = utils.get_fastx_read_number(f'{self.out_prefix}_{CANDIDATE_FQ_SUFFIX}')
        self.add_metric(
            name = 'Reads Mapped to Any V(D)J genes', 
            value = n_bcrtcr,
            total = self._matched_reads,
            help_info = "Fraction of reads that partially or wholly map to any germline V(D)J gene segment"
        )

        for _chain in self._chains:
            n_chain = utils.get_fastx_read_number(f'{self.out_prefix}_{_chain}.fq')
            self.add_metric(
                name = f'Reads Mapped to {_chain}', 
                value = n_chain, 
                total = self._matched_reads,
                help_info = f"Fraction of reads that map partially or wholly to a germline {_chain} gene segment."
            )

    def run(self):
        self.extract_chain_reads()
        self.add_metrics()


@utils.add_log
def mapping(args):
    with Mapping(args, display_title="Mapping") as runner:
        runner.run()


def get_opts_mapping(parser, sub_program):
    if sub_program:
        parser = s_common(parser)
        parser.add_argument('--match_fq2', help='R2 reads matched with scRNA-seq.', required=True)
        parser.add_argument('--match_fq1', help='R1 reads matched with scRNA-seq.', required=True)

    parser.add_argument('--ref', help='reference name', choices=["hg19", "hg38", "GRCm38", "other"], required=True)
    parser.add_argument('--seqtype', help='TCR/BCR seq data.', choices=['TCR', 'BCR'], required=True)
    parser.add_argument('--barcodeRange', help='Barcode range in fq1, INT INT CHAR.', default='0 23 +') 
    parser.add_argument('--umiRange', help='UMI range in fq1, INT INT CHAR.', default='24 -1 +')


from celescope.tools.tag.count_tag import Count_tag as Ct,  count_tag as ct, get_opts_count_tag as opts




class Count_tag(Ct):
    pass

def count_tag(args):
    ct(args)

def get_opts_count_tag(parser, sub_program):
    opts(parser, sub_program)


import pandas as pd

from celescope.tools.capture.analysis import Analysis, get_opts_analysis
from celescope.fusion.count_fusion import Count_fusion
from celescope.fusion.mkref import Mkref_fusion
from celescope.tools.plotly_plot import Tsne_dropdown_plot,Tsne_single_plot,Tsne_plot

def analysis_fusion(args):

    with Analysis_fusion(args) as runner:
        runner.run()


def get_opts_analysis_fusion(parser, sub_program):
    parser.add_argument('--fusion_genomeDir', help='Fusion genome directory.', required=True)
    get_opts_analysis(parser, sub_program)


class Analysis_fusion(Analysis):
    def __init__(self, args, display_title='Analysis'):
        super().__init__(args, display_title)

        fusion_pos_file =  Mkref_fusion.get_config(args.fusion_genomeDir)['files']['fusion_pos']
        self.pos_dict = Count_fusion.read_pos_file(fusion_pos_file)

        self.count_fusion_df = None
        self.count_fusion_out = f'{self.out_prefix}_fusion_count.csv'

    
    def plot_fusion(self):
        """
        plot fusion count
        """
        df_tsne_file = f'{self.out_prefix}_UMI_tsne.csv'
        df_tsne = pd.read_csv(df_tsne_file,sep=",",index_col=0)
        feature_name_list = df_tsne.columns[4:-1].to_list()
        # plot
        tsne_cluster = Tsne_plot(df_tsne, 'cluster').get_plotly_div()
        self.add_data(tsne_cluster=tsne_cluster)
        tsne_citeseq = Tsne_dropdown_plot(df_tsne,'Fusion',feature_name_list).get_plotly_div()
        self.add_data(tsne_citeseq=tsne_citeseq)
        Tsne_single_plot(df_tsne,feature_name_list,self.args.outdir).get_plotly_div()
    

    def get_fusion_count_df(self):
        fusion_df = self.df_tsne.reset_index().filter(list(self.pos_dict.keys()) + ["barcode"])
        fusion_df = fusion_df.melt(id_vars=["barcode"],value_name="UMI",var_name="fusion",ignore_index=True)
        
        count = fusion_df[fusion_df["UMI"] > 0].groupby(["fusion"], as_index=False)["barcode"].count()
        self.count_fusion_df = count.assign(percent = lambda x: 100 * x.barcode / len(self.df_tsne))
        self.count_fusion_df.to_csv(self.count_fusion_out, index = None)
    
    def add_fusion_count_metrics(self):
        self.add_help_content('fusion_1, fusion_2...', 'number of positive cells(UMI > 0) of each fusion')
        for _, row in self.count_fusion_df.iterrows():
            self.add_metric(
                name=f'{row["fusion"]}',
                value=int(row['barcode']),
                total=int(len(self.df_tsne)),
            )

    def run(self):
        super().run()
        self.plot_fusion()
        self.get_fusion_count_df()
        self.add_fusion_count_metrics()


import sys

import numpy as np
import scipy.stats as sp_stats


def determine_max_filtered_bcs(recovered_cells):
    """ Determine the max # of cellular barcodes to consider """
#     return float(recovered_cells) * cr_constants.FILTER_BARCODES_MAX_RECOVERED_CELLS_MULTIPLE # 6
    return float(recovered_cells) * 6


def init_barcode_filter_result():
    return {
        'filtered_bcs': 0,
        #'filtered_bcs_lb': 0,
        #'filtered_bcs_ub': 0,
        #'max_filtered_bcs': 0,
        'filtered_bcs_var': 0,
        #'filtered_bcs_cv': 0,
    }


def summarize_bootstrapped_top_n(top_n_boot):
    top_n_bcs_mean = np.mean(top_n_boot)
    #top_n_bcs_sd = np.std(top_n_boot)
    top_n_bcs_var = np.var(top_n_boot)
    result = {}
    result['filtered_bcs_var'] = top_n_bcs_var
    # comment these two lines out to avoid `ValueError: cannot convert float NaN to integer` when analyze data with low number of barcode
    #result['filtered_bcs_lb'] = round(sp_stats.norm.ppf(0.025, top_n_bcs_mean, top_n_bcs_sd))
    #result['filtered_bcs_ub'] = round(sp_stats.norm.ppf(0.975, top_n_bcs_mean, top_n_bcs_sd))
    result['filtered_bcs'] = int(round(top_n_bcs_mean))
    return result


def find_within_ordmag(x, baseline_idx):
    x_ascending = np.sort(x)
    baseline = x_ascending[-baseline_idx]
    cutoff = max(1, round(0.1 * baseline))
    # Return the index corresponding to the cutoff in descending order
    return len(x) - np.searchsorted(x_ascending, cutoff)


def filter_cellular_barcodes_ordmag(bc_counts, recovered_cells):
    """ Simply take all barcodes that are within an order of magnitude of a top barcode
        that likely represents a cell
    """
    if recovered_cells is None:
        # Modified parameter, didn't use the default value
        recovered_cells = 3000
#         recovered_cells = cr_constants.DEFAULT_RECOVERED_CELLS_PER_GEM_GROUP # 3000

    # Initialize filter result metrics
    metrics = init_barcode_filter_result()
    # determine max # of cellular barcodes to consider
    max_filtered_bcs = determine_max_filtered_bcs(recovered_cells)
    metrics['max_filtered_bcs'] = max_filtered_bcs

    nonzero_bc_counts = bc_counts[bc_counts > 0]
    if len(nonzero_bc_counts) == 0:
        msg = "WARNING: All barcodes do not have enough reads for ordmag, allowing no bcs through"
        return [], metrics, msg

    baseline_bc_idx = int(round(float(recovered_cells) * (1 - 0.99)))  # Quantile=0.99
    baseline_bc_idx = min(baseline_bc_idx, len(nonzero_bc_counts) - 1)
    assert baseline_bc_idx < max_filtered_bcs

    # Bootstrap sampling; run algo with many random samples of the data
    top_n_boot = np.array([
        find_within_ordmag(np.random.choice(nonzero_bc_counts, len(nonzero_bc_counts)), baseline_bc_idx)
        for i in range(100)  # 100
    ])

    metrics.update(summarize_bootstrapped_top_n(top_n_boot))

    # Get the filtered barcodes
    top_n = metrics['filtered_bcs']
    top_bc_idx = np.sort(np.argsort(bc_counts)[::-1][:top_n])
    return top_bc_idx, metrics, None


def filter_cellular_barcodes_fixed_cutoff(bc_counts, cutoff):
    nonzero_bcs = len(bc_counts[bc_counts > 0])
    top_n = min(cutoff, nonzero_bcs)
    # np.argsort(bc_counts) => the indices that would sort an array
    # np.argsort(bc_counts)[0] => idx of the smallest element in array
    # np.argsort(bc_counts)[-1] => idx of the largest element in array
    top_bc_idx = np.sort(np.argsort(bc_counts)[::-1][:top_n])
    metrics = {
        'filtered_bcs': top_n,
        'filtered_bcs_lb': top_n,
        'filtered_bcs_ub': top_n,
        'max_filtered_bcs': 0,
        'filtered_bcs_var': 0,
        'filtered_bcs_cv': 0,
    }
    return top_bc_idx, metrics, None



def est_background_profile_bottom(matrix, bottom_frac):
    """Construct a background expression profile from the barcodes that make up the bottom b% of the data
    Args:
      matrix (scipy.sparse.csc_matrix): Feature x Barcode matrix
      bottom_frac (float): Use barcodes making up the bottom x fraction of the counts (0-1)
    Returns:
      (nz_feat (ndarray(int)), profile_p (ndarray(float)): Indices of nonzero features and background profile
    """
    assert bottom_frac >= 0 and bottom_frac <= 1
    umis_per_bc = np.ravel(np.asarray(matrix.sum(0)))
    barcode_order = np.argsort(umis_per_bc)

    cum_frac = np.cumsum(umis_per_bc[barcode_order]) / float(umis_per_bc.sum())
    max_bg_idx = np.searchsorted(cum_frac, bottom_frac, side='left')
    bg_mat = matrix[:, barcode_order[0:max_bg_idx]]

    nz_feat = np.flatnonzero(np.asarray(bg_mat.sum(1)))
    bg_profile = np.ravel(bg_mat[nz_feat, :].sum(axis=1))
    bg_profile_p = bg_profile / float(np.sum(bg_profile))
    assert np.isclose(bg_profile_p.sum(), 1)

    return (nz_feat, bg_profile_p)


def eval_multinomial_loglikelihoods(matrix, profile_p, max_mem_gb=0.1):
    """Compute the multinomial log PMF for many barcodes
    Args:
      matrix (scipy.sparse.csc_matrix): Matrix of UMI counts (feature x barcode)
      profile_p (np.ndarray(float)): Multinomial probability vector
      max_mem_gb (float): Try to bound memory usage.
    Returns:
      log_likelihoods (np.ndarray(float)): Log-likelihood for each barcode
    """
    gb_per_bc = float(matrix.shape[0] * matrix.dtype.itemsize) / (1024**3)
    bcs_per_chunk = max(1, int(round(max_mem_gb/gb_per_bc)))
    num_bcs = matrix.shape[1]

    loglk = np.zeros(num_bcs)

    for chunk_start in range(0, num_bcs, bcs_per_chunk):
        chunk = slice(chunk_start, chunk_start+bcs_per_chunk)
        matrix_chunk = matrix[:, chunk].transpose().toarray()
        n = matrix_chunk.sum(1)
        loglk[chunk] = sp_stats.multinomial.logpmf(matrix_chunk, n, p=profile_p)
    return loglk


def simulate_multinomial_loglikelihoods(profile_p, umis_per_bc,
                                        num_sims=1000, jump=1000,
                                        n_sample_feature_block=1000000, verbose=False):
    """Simulate draws from a multinomial distribution for various values of N.
       Uses the approximation from Lun et al. ( https://www.biorxiv.org/content/biorxiv/early/2018/04/04/234872.full.pdf )
    Args:
      profile_p (np.ndarray(float)): Probability of observing each feature.
      umis_per_bc (np.ndarray(int)): UMI counts per barcode (multinomial N).
      num_sims (int): Number of simulations per distinct N value.
      jump (int): Vectorize the sampling if the gap between two distinct Ns exceeds this.
      n_sample_feature_block (int): Vectorize this many feature samplings at a time.
    Returns:
      (distinct_ns (np.ndarray(int)), log_likelihoods (np.ndarray(float)):
      distinct_ns is an array containing the distinct N values that were simulated.
      log_likelihoods is a len(distinct_ns) x num_sims matrix containing the
        simulated log likelihoods.
    """
    distinct_n = np.flatnonzero(np.bincount(umis_per_bc))

    loglk = np.zeros((len(distinct_n), num_sims), dtype=float)
    num_all_n = np.max(distinct_n) - np.min(distinct_n)
    if verbose:
        print('Number of distinct N supplied: %d' % len(distinct_n))
        print('Range of N: %d' % num_all_n)
        print('Number of features: %d' % len(profile_p))

    sampled_features = np.random.choice(len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True)
    k = 0

    log_profile_p = np.log(profile_p)

    for sim_idx in range(num_sims):
        if verbose and sim_idx % 100 == 99:
            sys.stdout.write('.')
            sys.stdout.flush()
        curr_counts = np.ravel(sp_stats.multinomial.rvs(distinct_n[0], profile_p, size=1))

        curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[0], p=profile_p)

        loglk[0, sim_idx] = curr_loglk

        for i in range(1, len(distinct_n)):
            step = distinct_n[i] - distinct_n[i-1]
            if step >= jump:
                # Instead of iterating for each n, sample the intermediate ns all at once
                curr_counts += np.ravel(sp_stats.multinomial.rvs(step, profile_p, size=1))
                curr_loglk = sp_stats.multinomial.logpmf(curr_counts, distinct_n[i], p=profile_p)
                assert not np.isnan(curr_loglk)
            else:
                # Iteratively sample between the two distinct values of n
                for n in range(distinct_n[i-1]+1, distinct_n[i]+1):
                    j = sampled_features[k]
                    k += 1
                    if k >= n_sample_feature_block:
                        # Amortize this operation
                        sampled_features = np.random.choice(
                            len(profile_p), size=n_sample_feature_block, p=profile_p, replace=True)
                        k = 0
                    curr_counts[j] += 1
                    curr_loglk += log_profile_p[j] + np.log(float(n)/curr_counts[j])

            loglk[i, sim_idx] = curr_loglk

    if verbose:
        sys.stdout.write('\n')

    return distinct_n, loglk


def compute_ambient_pvalues(umis_per_bc, obs_loglk, sim_n, sim_loglk):
    """Compute p-values for observed multinomial log-likelihoods
    Args:
      umis_per_bc (nd.array(int)): UMI counts per barcode
      obs_loglk (nd.array(float)): Observed log-likelihoods of each barcode deriving from an ambient profile
      sim_n (nd.array(int)): Multinomial N for simulated log-likelihoods
      sim_loglk (nd.array(float)): Simulated log-likelihoods of shape (len(sim_n), num_simulations)
    Returns:
      pvalues (nd.array(float)): p-values
    """
    assert len(umis_per_bc) == len(obs_loglk)
    assert sim_loglk.shape[0] == len(sim_n)

    # Find the index of the simulated N for each barcode
    sim_n_idx = np.searchsorted(sim_n, umis_per_bc)
    num_sims = sim_loglk.shape[1]

    num_barcodes = len(umis_per_bc)

    pvalues = np.zeros(num_barcodes)

    for i in range(num_barcodes):
        num_lower_loglk = np.sum(sim_loglk[sim_n_idx[i], :] < obs_loglk[i])
        pvalues[i] = float(1 + num_lower_loglk) / (1 + num_sims)
    return pvalues



"""
Integration tests
"""

import os
import subprocess
from concurrent import futures

from celescope.tools import utils

ASSAYS = [
    'fusion',
    'vdj',
    'tag',
    'capture_virus',
    'snp',
    'rna',
    'dynaseq',
]


def run_single(assay, test_dir):
    """
    Returns:
        string indicates complete status
    """
    os.chdir(os.path.join(test_dir, assay))
    print("*" * 20 + "running " + assay + "*" * 20)
    subprocess.check_call('sh run_shell.sh', shell=True)
    subprocess.check_call('sh sjm.sh', shell=True)
    try:
        subprocess.check_call('sh ./shell/test1.sh', shell=True)
    except subprocess.CalledProcessError:
        return f"{assay} failed"
    print("*" * 20 + "success " + assay + "*" * 20)
    return f"{assay} success."


@utils.add_log
def test_mutiple(assays, test_dir):
    """
    Run all
    >>> pytest -s celescope/tests/test_multi.py --test_dir {some_dir}
    Run some tests
    >>> pytest -s celescope/tests/test_multi.py --test_dir {some_dir} --assays tag,fusion
    """

    if not assays:
        assays = ASSAYS
    else:
        assays = assays.split(',')
    print("assays to run: ", assays)
    thread = len(assays)
    executor = futures.ProcessPoolExecutor(max_workers=thread)
    results = executor.map(run_single, assays, [test_dir] * len(assays))
    res_list = []
    for result in results:
        res_list.append(result)
    for result in res_list:
        print(result)
    assert not any((string.find("failed") != -1 for string in res_list))



import os
import subprocess
import sys

from celescope.tools import utils
from celescope.tools.make_ref import MakeRef, GENOME_CONFIG

class Mkref_snp:
    """
    ## Features
    - Create dictionary file and fasta index for gatk SplitNCigarReads.
    (https://gatk.broadinstitute.org/hc/en-us/articles/360035531652-FASTA-Reference-genome-format) 
    Need to run `celescope rna mkref` first

    ## Output
    - fasta index
    - gatk dictionary file

    ## Usage
    ```
    # run celescope rna mkref first
    celescope snp mkref \\
     --genome_name Homo_sapiens_ensembl_99 \\
     --fasta Homo_sapiens.GRCh38.dna.primary_assembly.fa
    ```
    """
    def __init__(self):
        pass

    @utils.add_log
    def build_fasta_index(self, fasta):
        cmd = (
            f'samtools faidx {fasta}'
        )
        sys.stderr.write(cmd + '\n')
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def build_fasta_dict(self, fasta):
        cmd = (
            f'gatk CreateSequenceDictionary -R {fasta}'
        )
        sys.stderr.write(cmd + '\n')
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def run(self):
        assert os.path.exists(GENOME_CONFIG), (
            f"{GENOME_CONFIG} not exist! "
            "You need to build rna genome with 'celescope rna mkref' first."
        )
        # reset genome type here
        config = MakeRef.get_config('./')
        fasta = config['files']['fasta']
        self.build_fasta_index(fasta)
        self.build_fasta_dict(fasta)

        config['meta']['genome_type'] = 'rna, snp'
        with open(GENOME_CONFIG, 'w') as config_handle:
            config.write(config_handle)
    

def mkref(args):
    # pylint: disable=unused-argument
    Mkref_snp().run()


def get_opts_mkref(parser, sub_program):
    # pylint: disable=unused-argument
    pass



from celescope.snp.__init__ import __ASSAY__
from celescope.tools.multi import Multi
from celescope.tools.__init__ import TAG_BAM_SUFFIX


class Multi_snp(Multi):
    """
    ## Usage

    ### Make a snp reference genomeDir

    1. Run `celescope rna mkref`. If you already have a rna genomeDir, you can use it and skip this step.
    2. Run `celescope snp mkref` under the rna genomeDir. Check [mkref.md](./mkref.md) for help.

    ### Run multi_snp
    There are two ways to run `multi_snp`

    1. Do not perform consensus before alignment and report read count(recommended for data generated with FocuSCOPE kit).

    ```
    multi_snp\\
        --mapfile ./test1.mapfile\\
        --genomeDir {genomeDir after running celescope snp mkref}\\
        --thread 4\\
        --mod shell\\
        --panel lung_1\\
        --not_consensus
    ```

    2. Do consensus before alignment and report UMI count. 

    ```
    multi_snp\\
        --mapfile ./test1.mapfile\\
        --genomeDir {genomeDir after running celescope snp mkref}\\
        --thread 4\\
        --mod shell\\
        --panel lung_1\\
    ```

    """

    def star(self, sample):
        step = 'star'
        cmd_line = self.get_cmd_line(step, sample)
        if self.args.not_consensus:
            fq = f'{self.outdir_dic[sample]["cutadapt"]}/{sample}_clean_2.fq{self.fq_suffix}'
        else:
            fq = f'{self.outdir_dic[sample]["consensus"]}/{sample}_consensus.fq'
            cmd_line += ' --consensus_fq '

        cmd = (
            f'{cmd_line} '
            f'--fq {fq} '
        )
        self.process_cmd(cmd, step, sample, m=self.args.starMem, x=self.args.thread)

    def target_metrics(self, sample):
        step = 'target_metrics'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["featureCounts"]}/{sample}_{TAG_BAM_SUFFIX}'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--add_RG '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)

    def variant_calling(self, sample):
        step = 'variant_calling'
        cmd_line = self.get_cmd_line(step, sample)
        bam = f'{self.outdir_dic[sample]["target_metrics"]}/{sample}_filtered_sorted.bam'
        cmd = (
            f'{cmd_line} '
            f'--bam {bam} '
            f'--match_dir {self.col4_dict[sample]} '
        )
        self.process_cmd(cmd, step, sample, m=8, x=1)

    def filter_snp(self, sample):
        step ='filter_snp'
        vcf = f'{self.outdir_dic[sample]["variant_calling"]}/{sample}_norm.vcf'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--vcf {vcf} '
        )
        self.process_cmd(cmd, step, sample, m=1, x=1)

    def analysis_snp(self, sample):
        step = 'analysis_snp'
        vcf = f'{self.outdir_dic[sample]["filter_snp"]}/{sample}_filtered.vcf'
        cmd_line = self.get_cmd_line(step, sample)
        cmd = (
            f'{cmd_line} '
            f'--match_dir {self.col4_dict[sample]} '
            f'--vcf {vcf} '
        )
        self.process_cmd(cmd, step, sample, m=2, x=1)


def main():
    multi = Multi_snp(__ASSAY__)
    multi.run()


if __name__ == '__main__':
    main()



from celescope.tools.tag.mapping_tag import Mapping_tag as Mt, mapping_tag as mt
from celescope.tools.step import s_common

# L23C15, not L25C15 like Clindex
def get_opts_mapping_tag(parser, sub_program):
    parser.add_argument(
        "--fq_pattern",
        help="""R2 read pattern. The number after the letter represents the number of bases. CLindex is `L25C15` and sweetseq is `L23C15`
`L` linker(common sequences)  
`C` tag barcode  
""",
        default='L23C15'
    )
    parser.add_argument(
        "--barcode_fasta",
        help="""Required. Tag barcode fasta file. It will check the mismatches between tag barcode 
sequence in R2 reads with all tag barcode sequence in barcode_fasta. 
It will assign read to the tag with mismatch < threshold. 
If no such tag exists, the read is classified as invalid.

You can find the barcode fasta file under `celescope/data/sweetseq`
""",
        required=True,
    )
    parser.add_argument(
        "--linker_fasta",
        help="""Optional. If provided, it will check the mismatches between linker sequence in R2 reads 
with all linker sequence in linker_fasta. If no mismatch < len(linker) / 10 + 1, the read is classified as invalid.
""",
    )
    if sub_program:
        s_common(parser)
        parser.add_argument("--fq", help="R2 read fastq.", required=True)

class Mapping_tag(Mt):
    pass

def mapping_tag(args):
    mt(args)


import json
import collections

import pandas as pd
import pysam

from celescope.tools import utils
from celescope.tools.step import Step, s_common
from celescope.__init__ import HELP_DICT

class FeatureCounts(Step):
    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        self.filter_bam = f'{self.outdir}/{self.sample}_filter.bam'
        self.coord_sorted_bam = f'{self.outdir}/{self.sample}_filter.bam.featureCounts.bam'
        self.name_sorted_bam = f'{self.outdir}/{self.sample}_filter_name_sorted.bam'
        self.log_file = f'{self.outdir}/{self.sample}.summary'

    @staticmethod
    def get_valid_barcodes(filter_umi_file):
        df = pd.read_csv(filter_umi_file)
        df = df[df['sum_UMI'] > 0]
        valid_barcodes = set(df['barcode'])
        return valid_barcodes

    @staticmethod
    def get_valid_umis(filter_read_count_json, valid_barcodes):
        barcode_umis = collections.defaultdict(set)
        with open(filter_read_count_json) as fh:
            dic = json.load(fh)
        for barcode in dic:
            if barcode in valid_barcodes:
                for ref in dic[barcode]:
                    for umi in dic[barcode][ref]:
                        if dic[barcode][ref][umi] > 0:
                            barcode_umis[barcode].add(umi)
        return barcode_umis

    def run_filter(self):
        valid_barcodes = FeatureCounts.get_valid_barcodes(self.args.filter_umi_file)
        barcode_umis = FeatureCounts.get_valid_umis(self.args.filter_read_count_json, valid_barcodes)

        with pysam.AlignmentFile(self.args.bam, "rb") as raw_bam:
            header = raw_bam.header
            with pysam.AlignmentFile(self.filter_bam, "w", header=header) as filter_bam:
                for read in raw_bam:
                    attr = read.query_name.split(':')
                    barcode = attr[0]
                    umi = attr[1]
                    if barcode in valid_barcodes and umi in barcode_umis[barcode]:
                        filter_bam.write(read)         
    
    @utils.add_log
    def run_featureCounts(self):
        cmd = (
            'featureCounts '
            '-s 1 '
            f'-a {self.args.gtf} '
            f'-o {self.outdir}/{self.sample} '  # not bam
            '-R BAM '
            f'-T {self.thread} '
            f'-t gene '
            f'{self.filter_bam} '
            '2>&1 '
        )
        if self.args.featureCounts_param:
            cmd += (" " + self.args.featureCounts_param)
        self.debug_subprocess_call(cmd)

        utils.sort_bam(
            input_bam = self.coord_sorted_bam,
            output_bam = self.name_sorted_bam,
            by='name'
        )

    @staticmethod
    def read_log(log_file):
        """
        Args:
            log_file: featureCounts log summary file
        Returns:
            log_dict: {'Assigned': 123, ...}
        """
        # skip first line
        df = pd.read_csv(log_file, sep='\t', header=None, names=['name', 'value'], skiprows=1)
        log_dict = df.set_index('name')['value'].to_dict()
        return log_dict

    def add_metrics(self):
        log_dict = self.read_log(self.log_file)
        total = sum(log_dict .values())
        self.add_metric(
            name='Reads Assigned',
            value=log_dict['Assigned'],
            total=total,
            help_info='Reads that can be successfully assigned'
        )

        self.add_metric(
            name='Reads Unassigned_NoFeatures',
            value=log_dict['Unassigned_NoFeatures'],
            total=total,
            help_info='Reads that can not be ssigned'
        )

        self.add_metric(
            name='Reads Unassigned_Ambiguity',
            value=log_dict['Unassigned_Ambiguity'],
            total=total,
            help_info='Reads that overlap two or more features'
        )


    def run(self):
        self.run_filter()
        self.run_featureCounts()
        self.add_metrics()

@utils.add_log
def featureCounts(args):
    if args.gtf:
        with FeatureCounts(args) as runner:
            runner.run()


def get_opts_featureCounts(parser, sub_program):
    parser.add_argument(
        "--gtf",
        help="Optional. Genome gtf file. Use absolute path or relative path to `genomeDir`.",
        )
    if sub_program:
        parser.add_argument('--bam', help='input bam file', required=True)
        parser.add_argument('--filter_umi_file', help='filter umi file', required=True)
        parser.add_argument('--filter_read_count_json', help='filter_read_count_json file', required=True)
        parser.add_argument('--featureCounts_param', help=HELP_DICT['additional_param'], default="")
        s_common(parser)


from celescope.tools.make_ref import MakeRef_STAR

def mkref(args):
    genome_type = 'virus'
    with MakeRef_STAR(genome_type, args) as runner:
        runner.run()

def get_opts_mkref(parser, sub_program):
    MakeRef_STAR.opts(parser, sub_program)



import argparse
import glob
import os


class Debug():
    def __init__(self):
        self.outdir = 'debug'
        if not os.path.exists(self.outdir):
            os.mkdir(self.outdir)
        self.opts()
        self.all_reads = glob.glob(f'{self.dir}/02.cutadapt/*clean_2.fq.gz')[0]
        # out files
        self.sub_reads = f'{self.outdir}/sub.fq.gz'
        self.unmapped = f'{self.outdir}/sub_Unmapped.out.mate1'

    def opts(self):
        readme = 'debug'
        parser = argparse.ArgumentParser(readme)
        parser.add_argument('--dir', required=True)
        parser.add_argument('--n_sub', default=1000000)
        parser.add_argument('--genomeDir', default='/SGRNJ/Public/Database/genome/homo_sapiens/ensembl_92')
        args = parser.parse_args()
        self.dir = args.dir
        self.n_sub = args.n_sub
        self.genomeDir = args.genomeDir
        self.thread = 2

    def run_subsample(self):
        cmd = (
            f'less {self.all_reads} '
            f'|head -n {self.n_sub * 4} | gzip -c > {self.sub_reads} '
        )
        with open('subsample.sh', 'wt') as f:
            f.write(cmd)

    def run_STAR(self):
        cmd = (
            f'star --runThreadN {self.thread} \\\n'
            f'--genomeDir {self.genomeDir} \\\n'
            f'--readFilesIn {self.sub_reads} \\\n'
            f'--readFilesCommand zcat --outFilterMultimapNmax 1 \\\n'
            f'--outFileNamePrefix {self.outdir}/sub_ \\\n'
            f'--outSAMtype BAM SortedByCoordinate --outFilterMatchNmin 0 \\\n'
            f'--outReadsUnmapped Fastx \\\n'
        )
        with open('STAR.sh', 'wt') as f:
            f.write(cmd)

    def fastqc(self):
        cmd = (
            f'fastqc {self.unmapped} \\\n'
        )
        with open('fastqc.sh', 'wt') as f:
            f.write(cmd)

    def run(self):
        self.run_subsample()
        self.run_STAR()
        self.fastqc()


if __name__ == '__main__':
    de = Debug()
    de.run()



'''
bulk-seq vdj mapping
'''
import pandas as pd
import subprocess

from celescope.tools import utils
from celescope.vdj import mapping_vdj as super_vdj


class Mapping_vdj(super_vdj.Mapping_vdj):
    """
    ## Features
    - Align R2 reads to IMGT(http://www.imgt.org/) database sequences with blast.
    ## Output
    - `{sample}_airr.tsv` The alignment result of each read.
    A tab-delimited file compliant with the AIRR Rearrangement schema(https://docs.airr-community.org/en/stable/datarep/rearrangements.html)
    - `{sample}_produtive.tsv` Including all productive chains.
    """
    def __init__(self, args, display_title=None):
        super().__init__(args, display_title=display_title)

        # out
        self.airr_out = f'{self.out_prefix}_airr.tsv'
        self.productive_file = f'{self.out_prefix}_productive.tsv'

    @utils.add_log
    def igblast(self, chain, ig_seqtype):
        cmd = (
            f"igblastn -query {self.args.fasta} "
            f"-organism {self.species} "
            f"-ig_seqtype {ig_seqtype} "
            f"-auxiliary_data optional_file/{self.species}_gl.aux "
            f"-num_threads {self.args.thread} "
            f"-germline_db_V {self.ref_path}/{chain}V.fa "
            f"-germline_db_D {self.ref_path}/{chain}D.fa "
            f"-germline_db_J {self.ref_path}/{chain}J.fa "
            "-domain_system imgt -show_translation -outfmt 19 " # outfmt19 is an AIRR tab-delimited file, IgBLAST v1.9.0 or higher required.
            f"-out {self.airr_out} "
        )

        self.igblast.logger.info(cmd)
        subprocess.check_call(cmd, shell=True)

    @utils.add_log
    def mapping_summary(self):
        df = pd.read_csv(self.airr_out, sep='\t')
        df.fillna("", inplace=True)
        total_reads = df.shape[0]

        self.add_metric(
            name="Species",
            value=self.species,
            help_info="Human or Mouse"
        )
        
        df = df[(df["v_call"]!="") | ((df["d_call"]!="")) | ((df["j_call"]!=""))]
        self.add_metric(
            name="UMIs Mapped to Any VDJ Gene",
            value=df.shape[0],
            total=total_reads,
            help_info="UMIs Mapped to any germline VDJ gene segments"
        )

        # Reads with CDR3
        df_cdr3 = df[(df["cdr3_aa"]!="") & (df["junction_aa"]!="")]
        self.add_metric(
            name="UMIs with CDR3",
            value=df_cdr3.shape[0],
            total=total_reads,
            help_info="UMIs with CDR3 sequence"
        )

        # Reads with Correct CDR3
        df_correct_cdr3 = df_cdr3[~(df_cdr3["cdr3_aa"].str.contains(r"\*")) & ~(df_cdr3["cdr3_aa"].str.contains("X"))]
        self.add_metric(
            name="UMIs with Correct CDR3",
            value=df_correct_cdr3.shape[0],
            total=total_reads,
            help_info="UMIs with CDR3 might have stop codon and these Reads are classified as incorrect"
        )
        
        # Reads Mapped Confidently To VJ Gene
        df_confident = df_correct_cdr3[df_correct_cdr3["productive"]=="T"]
        self.add_metric(
            name="UMIs Mapped Confidently to VJ Gene",
            value=df_confident.shape[0],
            total=total_reads,
            help_info="UMIs with productive rearrangement mapped to VJ gene pairs and with correct CDR3"
        )

        # Reads Mapped Confidently to each chain
        for chain in self.chains:
            df_chain = df_confident[df_confident.locus == chain]
            self.add_metric(
                name=f"UMIs Mapped to {chain}",
                value=df_chain.shape[0],
                total=total_reads,
                help_info=f"UMIs mapped confidently to {chain}"
            )
        
        # output file
        df_confident["barcode"] = df_confident["sequence_id"].apply(lambda x: x.split(':')[0])
        df_VJ = df_confident[["barcode", "sequence_id","locus", "v_call", "d_call", "j_call", "junction", "junction_aa"]]
        df_VJ.rename(columns={
            "locus": "chain",
            "v_call": "bestVGene",
            "d_call": "bestDGene",
            "j_call": "bestJGene",
            "junction": "nSeqCDR3",
            "junction_aa": "aaSeqCDR3"
        }, inplace=True)

        for i in ["bestVGene", "bestDGene", "bestJGene"]:
            df_VJ[i] = df_VJ[i].apply(lambda x: x.split("*")[0])

        # CDR3 sequence have at least 5 amino acids and start with C
        df_VJ = df_VJ[df_VJ["aaSeqCDR3"].str.len()>5]
        df_VJ = df_VJ[df_VJ["aaSeqCDR3"].str.startswith('C')]
        df_VJ.to_csv(self.productive_file, sep="\t", index=False)

    def run(self):
        # run igblstn
        if self.seqtype == "TCR":
            self.igblast(chain="TR", ig_seqtype="TCR")
        else:
            self.igblast(chain="IG", ig_seqtype="Ig")
        self.mapping_summary()


def mapping_vdj(args):
    with Mapping_vdj(args, display_title="Mapping") as runner:
        runner.run()


def get_opts_mapping_vdj(parser, sub_program):
    super_vdj.get_opts_mapping_vdj(parser, sub_program)


import argparse
import pysam


def main():

    parser = argparse.ArgumentParser('Add barcode and UMI tag to celescope BAM')
    parser.add_argument('--bam', help='input bam', required=True)
    parser.add_argument('--outdir', help='output directory', default='./')
    args = parser.parse_args()
    add_tag(args.bam, args.outdir)


def add_tag(bam, outdir):

    bam_out = f"{outdir}/add_tag.bam"
    handle_in = pysam.AlignmentFile(bam, "rb")
    header = handle_in.header
    handle_out = pysam.AlignmentFile(bam_out, "wb", header=header)
    for read in handle_in:
        attr = read.query_name.split(':')
        barcode = attr[0]
        umi = attr[1]
        read.set_tag(tag='CB', value=barcode, value_type='Z')
        read.set_tag(tag='UB', value=umi, value_type='Z')
        handle_out.write(read)
    handle_in.close()
    handle_out.close()


if __name__ == '__main__':
    main()


import numpy as np
import pandas as pd

from celescope.tools import utils
from celescope.__init__ import HELP_DICT
from celescope.tools.step import Step, s_common
from celescope.vdj.__init__ import CHAINS, PAIRS
from celescope.tools.capture.threshold import Auto
from celescope.tools.emptydrop_cr import get_plot_elements


# Const
SEQUENCES_HEADER = ["aaSeqCDR3", "nSeqCDR3"]
# Auto filter percentitle
PERCENTILE = {
    'TCR': 85,
    'BCR': 90,
}


def target_cell_calling(df_UMI_sum, expected_target_cell_num=3000, target_barcodes=None, weight=3, coef=10, 
    percentile=99, UMI_min=3, umi_col='UMI'):
    """
    Args:
        df_UMI_sum: A dataframe with columns barcode and UMI. The barcode are match cell barcodes.
    
    Returns:
        umi_threhold: int
        target_cell_barcodes: list

    >>> df_UMI_sum = pd.DataFrame({"barcode": ["A", "B", "C", "D", "E"], "UMI": [1, 2, 1, 30, 40]})
    >>> umi_threshold, target_cell_barcodes = target_cell_calling(df_UMI_sum, expected_target_cell_num=5, percentile=80, coef=10, target_barcodes=["A", "C"])
    >>> umi_threshold == 3
    True
    >>> target_cell_barcodes == {'A', 'C', 'D', 'E'}
    True
    """
    umi_threshold = Auto(list(df_UMI_sum[umi_col]), expected_cell_num=expected_target_cell_num, coef=coef, percentile=percentile).run()
    if umi_threshold < UMI_min:
        umi_threshold = UMI_min
    # avoid change the original dataframe
    df_temp = df_UMI_sum.copy()
    if target_barcodes:
        df_temp[umi_col] = df_temp.apply(
            lambda row:  row[umi_col] * weight if row['barcode'] in target_barcodes else row[umi_col], axis=1)
             
    target_cell_barcodes = set(df_temp.loc[df_temp[umi_col] >= umi_threshold].barcode)

    return umi_threshold, target_cell_barcodes


class Count_vdj(Step):
    """
    ## Features
    - Cell-calling based on barcode-UMI rank.    
    - Summarize clonetypes infomation.

    ## Output
    - `{sample}_cell_confident.tsv` The clone type of VDJ cell barcode, each chain occupies one line.

    - `{sample}_cell_confident_count.tsv` The clone type of VDJ cell barcode, each cell occupies one line.

    - `{sample}_clonetypes.tsv` The count and percentage of each clonetypes of VDJ cell barcode.

    - `{sample}_match_clonetypes.tsv` When summarize clonetypes, only consider barcodes in the match scRNA-Seq library. 
    This file will only be produced when the `match_dir` parameter is provided.
    """

    def __init__(self, args, display_title=None):
        Step.__init__(self, args, display_title=display_title)

        # set
        self.chains = CHAINS[args.type]
        self.pairs = PAIRS[args.type]
        self.cols = []
        for chain in self.chains:
            for seq in SEQUENCES_HEADER:
                self.cols.append("_".join([seq, chain]))

        if utils.check_arg_not_none(self.args, 'match_dir'):
            self.match_cell_barcodes, _match_cell_number = utils.get_barcode_from_match_dir(args.match_dir)
        elif utils.check_arg_not_none(self.args, 'matrix_dir'):
            self.match_cell_barcodes, _match_cell_number = utils.get_barcode_from_matrix_dir(args.matrix_dir)
        else:
            raise FileNotFoundError("--match_dir or --matrix_dir is required.")
        self.match_cell_barcodes = set(self.match_cell_barcodes)

        # input data
        df_UMI_count_filter = pd.read_csv(args.UMI_count_filter_file, sep='\t')
        self.df_UMI_sum = df_UMI_count_filter.groupby(
            ['barcode'], as_index=False).agg({"UMI": "sum"})
        self.df_match_UMI_count_filter = df_UMI_count_filter[df_UMI_count_filter['barcode'].isin(self.match_cell_barcodes)]
        self.df_match_UMI_sum = self.df_UMI_sum[self.df_UMI_sum['barcode'].isin(self.match_cell_barcodes)]

        if args.target_cell_barcode:
            self.target_barcodes = utils.read_one_col(args.target_cell_barcode)
            self.expected_target_cell_num = len(self.target_barcodes)
        else:
            self.target_barcodes = None
            self.expected_target_cell_num = args.expected_target_cell_num


        # out files
        self.UMI_sum_file = f"{self.out_prefix}_UMI_sum.tsv"
        self.cell_confident_file = f"{self.out_prefix}_cell_confident.tsv"
        self.cell_confident_count_file = f"{self.out_prefix}_cell_confident_count.tsv"
        self.clonetypes_file = f"{self.out_prefix}_clonetypes.tsv"
        self.match_clonetypes_file = f"{self.out_prefix}_match_clonetypes.tsv"
        self.filtered_annotation = f"{self.out_prefix}_filtered_contig_annotations.csv"


    @utils.add_log
    def cell_calling(self):
        """

        """
        # UMI_min filter
        percentile = PERCENTILE[self.args.type]
        umi_threshold, target_cell_barcodes = target_cell_calling(
            self.df_match_UMI_sum, 
            expected_target_cell_num=self.expected_target_cell_num, 
            target_barcodes=self.target_barcodes,
            UMI_min=self.args.UMI_min,
            percentile=percentile
        )
        df_cell = self.df_match_UMI_count_filter[self.df_match_UMI_count_filter['barcode'].isin(
            target_cell_barcodes)]
        self.df_UMI_sum['mark'] = self.df_UMI_sum['barcode'].apply(lambda x: 'CB' if x in target_cell_barcodes else 'UB')
        self.df_UMI_sum = self.df_UMI_sum.sort_values(by=['UMI'], ascending=False)
        self.df_UMI_sum.to_csv(self.UMI_sum_file, sep='\t', index=False)
        self.add_data(chart=get_plot_elements.plot_barcode_rank(self.UMI_sum_file))

        self.add_metric(
            name="Estimated Number of Cells",
            value=len(target_cell_barcodes),
            help_info=f"number of barcodes considered as cell-associated. \
                UMI_threshold: {umi_threshold}. expected_cell_num: {self.expected_target_cell_num}"
        )

        return df_cell, target_cell_barcodes

    @utils.add_log
    def get_df_confident(self, df_cell):
        """
        1. UMI > iUMI
        2. in chain
        """
        if self.args.type == 'TCR':
            iUMI = self.args.TCR_iUMI
        elif self.args.type == 'BCR':
            iUMI = self.args.BCR_iUMI
        df_iUMI = df_cell[df_cell.UMI >= iUMI]
        df_confident = df_iUMI[df_iUMI["chain"].isin(self.chains)]
        df_confident = df_confident.sort_values("UMI", ascending=False)
        df_confident = df_confident.groupby(
            ["barcode", "chain"], as_index=False).head(1)
        return df_confident

    def get_df_valid_count(self, df_confident):
        df_valid_count = df_confident.set_index(["barcode", "chain"])
        df_valid_count = df_valid_count.unstack()
        df_valid_count.columns = ['_'.join(col) for col in df_valid_count]
        df_valid_count = df_valid_count.reset_index()
        df_valid_count.fillna(inplace=True, value="NA")
        return df_valid_count

    def get_clonetypes_and_write(self, df_valid_count):
        """
        Returns
        - df_clonetypes
        """

        df_clonetypes = df_valid_count.copy()
        self.cols = [x for x in self.cols if x in df_clonetypes.columns]
        df_clonetypes = df_clonetypes.groupby(self.cols, as_index=False).agg({
            "barcode": "count"})
        # put na last
        df_clonetypes.replace('NA', np.nan, inplace=True)
        df_clonetypes.sort_values(
            ["barcode"] + self.cols, ascending=False, na_position='last', inplace=True)
        df_clonetypes.replace(np.nan, 'NA', inplace=True)

        total_CDR3_barcode_number = sum(df_clonetypes.barcode)
        df_clonetypes["percent"] = df_clonetypes.barcode / \
            total_CDR3_barcode_number * 100
        df_clonetypes["percent"] = df_clonetypes["percent"].apply(
            lambda x: round(x, 2))

        # add clonetype ID
        df_clonetypes = df_clonetypes.reset_index()
        df_clonetypes["clonetype_ID"] = pd.Series(df_clonetypes.index) + 1
        df_clonetypes.drop(columns=["index"], inplace=True)

        # order
        order = ["clonetype_ID"] + self.cols + ["barcode", "percent"]
        df_clonetypes = df_clonetypes[order]
        df_clonetypes.rename(
            columns={"barcode": "barcode_count"}, inplace=True)
        # out clonetypes
        df_clonetypes.to_csv(self.clonetypes_file, sep="\t", index=False)

        return df_clonetypes

    def add_metrics(self, df_valid_count, cell_barcodes):
        n_cell = len(cell_barcodes)

        for chain in self.chains:
            UMI_col_name = "UMI_" + chain
            if UMI_col_name in df_valid_count.columns:
                df_valid_count[UMI_col_name].replace("NA", 0, inplace=True)
                Mean_chain_UMIs_per_Cell = round(np.mean(df_valid_count[UMI_col_name]),1)

            self.add_metric(
                name=f"Mean {chain} UMIs per Cell",
                value=Mean_chain_UMIs_per_Cell,
                help_info=f"mean number of UMI mapped to {chain} per cell"
            )

        if self.args.type == 'TCR':
            iUMI = self.args.TCR_iUMI
        elif self.args.type == 'BCR':
            iUMI = self.args.BCR_iUMI

        for pair in self.pairs:
            df_pair = df_valid_count.copy()
            for chain in pair:
                cdr3_col_name = "aaSeqCDR3_" + chain
                if cdr3_col_name not in df_pair.columns:
                    df_pair = pd.DataFrame({'barcode':[]})
                    break
                else:
                    df_pair = df_pair[df_pair[cdr3_col_name] != "NA"]
            
            n_cell_pair = len(df_pair.barcode.unique())

            pair_str = ','.join(pair)
            self.add_metric(
                name=f"Cells with ({pair_str}) pair",
                value=n_cell_pair,
                total=n_cell,
                help_info=f"cells with as least {iUMI} UMI mapped to each chain"
            )

    def write_cell_confident_count(self, df_valid_count, df_clonetypes, df_confident):
        df_mergeID = pd.merge(df_valid_count,
                              df_clonetypes, how="left", on=self.cols)
        df_mergeID.sort_values(["clonetype_ID", "barcode"], inplace=True)
        # output df_valid_count
        df_mergeID.to_csv(self.cell_confident_count_file,
                          sep="\t", index=False)
        df_mergeID = df_mergeID[["barcode", "clonetype_ID"]]
        df_cell_confident_with_ID = pd.merge(
            df_confident, df_mergeID, how="left", on="barcode")
        df_cell_confident_with_ID.sort_values(
            ["clonetype_ID", "barcode", "chain"], inplace=True)
        # output df_cell_confident
        df_cell_confident_with_ID.to_csv(
            self.cell_confident_file, sep="\t", index=False)

    def write_clonetypes_table_to_data(self, df_clonetypes):
        # cloneytpes table
        def format_table(df_clonetypes):
            df_table = df_clonetypes.copy()
            df_table["percent"] = df_table["percent"].apply(
                lambda x: str(x) + "%")
            seqs = ["aaSeqCDR3"]
            cols = []
            for chain in self.chains:
                for seq in seqs:
                    cols.append("_".join([seq, chain]))
            df_table_cols = ["clonetype_ID"] + \
                cols + ["barcode_count", "percent"]
            df_table_cols = [x for x in df_table_cols if x in df_table.columns]
            df_table = df_table[df_table_cols]
            table_header = ["Clonetype_ID"] + cols + ["Frequency", "Percent"]
            return df_table, table_header

        df_table, _table_header = format_table(df_clonetypes)
        title = 'Clonetypes'

        table_dict = self.get_table_dict(
            title=title,
            table_id='clonetypes',
            df_table=df_table
        )
        self.add_data(table_dict=table_dict)
    
    @utils.add_log
    def convert_to_annotation(self):
        in_file = pd.read_csv(self.cell_confident_file, sep='\t')
        out_file = pd.DataFrame({
                        'barcode':in_file.barcode,
                        'chain':in_file.chain,
                        'v_gene':in_file.bestVGene,
                        'd_gene':'None',
                        'j_gene':in_file.bestJGene,
                        'c_gene':'None',
                        'full_length':True,
                        'productive':True,
                        'cdr3':in_file.aaSeqCDR3,
                        'cdr3_nt':in_file.nSeqCDR3,
                        'reads':in_file.UMI,
                        'umis':in_file.UMI,
                        'raw_clonotype_id':in_file.clonetype_ID
                        })
        out_file.to_csv(self.filtered_annotation, sep=',', index=False)

    def run(self):
        df_cell, cell_barcodes = self.cell_calling()
        df_confident = self.get_df_confident(df_cell)
        df_valid_count = self.get_df_valid_count(df_confident)
        df_clonetypes= self.get_clonetypes_and_write(
            df_valid_count)
        self.add_metrics(df_valid_count, cell_barcodes)
        self.write_cell_confident_count(
            df_valid_count, df_clonetypes, df_confident)
        self.write_clonetypes_table_to_data(df_clonetypes)
        self.convert_to_annotation()


def count_vdj(args):
    with Count_vdj(args, display_title="Count") as runner:
        runner.run()


def get_opts_count_vdj(parser, sub_program):
    parser.add_argument(
        "--type", help="Required. `TCR` or `BCR`. ", required=True)
    parser.add_argument(
        '--UMI_min',
        help='minimum number of chain UMI to consider as as cell',
        type=int,
        default=3,
    )
    parser.add_argument(
        '--BCR_iUMI',
        help="""Minimum number of UMI of identical receptor type and CDR3 for BCR. 
For each (barcode, chain) combination, only UMI>=iUMI is considered valid.""",
        type=int,
        default=2,
    )
    parser.add_argument(
        '--TCR_iUMI',
        help="""Minimum number of UMI of identical receptor type and CDR3 for TCR. 
For each (barcode, chain) combination, only UMI>=iUMI is considered valid.""",
        type=int,
        default=1,
    )
    parser.add_argument(
        "--expected_target_cell_num", 
        help="Expected T or B cell number. If `--target_cell_barcode` is provided, this argument is ignored.", 
        type=int,
        default=3000,
    )
    parser.add_argument(
        "--target_cell_barcode", 
        help="Barcode of target cells. It is a plain text file with one barcode per line. If provided, `--expected_target_cell_num` is ignored.",
        default=None,
    )
    parser.add_argument(
        "--target_weight", 
        help="UMIs of the target cells are multiplied by this factor. Only used when `--target_cell_barcode` is provided.", 
        type=float,
        default=3.0,
    )
    if sub_program:
        parser.add_argument("--UMI_count_filter_file",
                            help="Required. File from step mapping_vdj.", required=True)
        parser.add_argument("--match_dir", help=HELP_DICT['match_dir'])
        parser.add_argument("--matrix_dir", help=HELP_DICT['matrix_dir'])

        parser = s_common(parser)
